#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for in-core sketched triangular estimation."""


import matplotlib.pyplot as plt
import pytest
import torch

from skerch.synthmat import SynthMat
from skerch.triangles import TriangularLinOp, serrated_hadamard_pattern
from skerch.utils import BadShapeError, gaussian_noise

from . import rng_seeds, torch_devices  # noqa: F401


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def serrated_hadamard_dims_chunks():
    """Pairs of ``(matrix_order, step_size)`` to test serrated Hadamard."""
    result = [(100, 2), (100, 3), (100, 10), (100, 100)]
    return result


@pytest.fixture
def hadamard_atol():
    """Absolute tolerances to test serrated Hadamard."""
    result = {torch.float64: 1e-13, torch.float32: 1e-5}
    return result


@pytest.fixture
def num_tri_tests(request):
    """Number of random vectors to test for each triangular linop case."""
    result = 3
    if request.config.getoption("--quick"):
        result = 1
    return result


@pytest.fixture
def dim_rank_decay_sym_width_hutch_maindiag_rtol(request):
    """Test cases for triangular estimation.

    Entries are in the form
    ``(dim, rank, decay, sym, width, hutch, maindiag, rtol)``, where:

    * ``dim`` is the number of dimensions of a square test matrix of
      exp-decaying singular values
    * ``rank`` is the number of unit singular values in the test matrix
    * ``decay`` is the speed of the exp-decay of singular values after ``rank``.
      The larger, the faster decay.
    * ``sym`` is a boolean specifying whether the matrix is symmetric
    * ``width`` is the step width for the staircase exact measurements in the
      triangular linop.
    * ``hutch`` is the number of Hutchinson measurements for the "serrated"
      (block-diagonal) estimator in the triangular linop.
    * ``maindiag`` is a boolean specifying whether to include the main diagonal
      in the triangle or not.
    * ``rtol`` is the relative error tolerance for each given case.
    """
    dims, rank = 1000, 100
    if request.config.getoption("--quick"):
        dims, rank = 500, 50
    result = [
        # [fast decay, asym]: just step measurements do decently
        (dims, rank, 0.5, False, 5, 0, True, 0.15),
        # ... but adding Hutch helps
        (dims, rank, 0.5, False, 5, 300, True, 0.1),
        # [fast decay, sym]: stronger diag, just steps is much worse now
        (dims, rank, 0.5, True, 5, 0, True, 0.5),
        # Hutch helps a lot, but it is recommended to deflate diagonal.
        (dims, rank, 0.5, True, 5, 300, True, 0.25),
        # Not including the diagonal shows that this was indeed the culprit
        (dims, rank, 0.5, True, 5, 100, False, 0.1),
        # [slow decay]: behave same as slow decay
        (dims, rank, 0.01, False, 5, 0, True, 0.15),
        (dims, rank, 0.01, True, 5, 100, False, 0.1),
    ]
    return result


# ##############################################################################
# # TRIANGULAR LINOP
# ##############################################################################
def test_triangular_linop_correctness(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    dim_rank_decay_sym_width_hutch_maindiag_rtol,
    num_tri_tests,
):
    """Test case for approximated triangulars on larger matrices.

    See :func:`dim_rank_decay_sym_width_hutch_maindiag_rtol` docstring and
    implementation for details on the use cases.
    """
    for seed in rng_seeds:
        for dtype in (torch.float32,):  # speed up test by just running low res
            for device in torch_devices:
                for (
                    dim,
                    rank,
                    decay,
                    sym,
                    stair_width,
                    hutch_meas,
                    maindiag,
                    rtol,
                ) in dim_rank_decay_sym_width_hutch_maindiag_rtol:
                    # create a linop for this seed/dtype/device and test case
                    # below, we will loop over lower/upper, left/right matmul
                    mat = SynthMat.exp_decay(
                        shape=(dim, dim),
                        rank=rank,
                        decay=decay,
                        symmetric=sym,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                        psd=False,
                    )
                    for lower in (True, False):
                        # create exact triangular and its approximation
                        diag_idx = int(not maindiag)  # if with main, idx=0
                        if lower:
                            tri = torch.tril(mat, diagonal=-diag_idx)
                        else:
                            tri = torch.triu(mat, diagonal=diag_idx)
                        tri_approx = TriangularLinOp(
                            mat,
                            stair_width,
                            hutch_meas,
                            lower=lower,
                            with_main_diagonal=maindiag,
                        )
                        # do the matmul, and check that error is within rtol
                        for i in range(num_tri_tests):
                            v = gaussian_noise(
                                dim,
                                seed=seed + i + 1,
                                dtype=dtype,
                                device=device,
                            )
                            for adjoint in (True, False):
                                if adjoint:
                                    tri_v = v @ tri
                                    tri_approx_v = v @ tri_approx
                                else:
                                    tri_v = tri @ v
                                    tri_approx_v = tri_approx @ v
                                #
                                dist = torch.dist(tri_v, tri_approx_v)
                                rel_err = (dist / torch.norm(tri_v)).item()
                                assert (
                                    rel_err <= rtol
                                ), "Incorrect triangular recovery!"


def test_triangular_linop_badshape():
    """Test case for non-square or empty inputs to triangular linop."""
    # non-square case
    mat = torch.ones(2, 3)
    with pytest.raises(BadShapeError):
        TriangularLinOp(
            mat,
            1,
            0,
            lower=True,
            with_main_diagonal=True,
        )
    # empty case
    mat = torch.ones(0, 0)
    with pytest.raises(BadShapeError):
        TriangularLinOp(
            mat,
            0,
            0,
            lower=True,
            with_main_diagonal=True,
        )


def test_triangular_linop_small_shapes():
    """Exact test cases for matrices of order 1, 2, 3 (both step and Hutch)."""
    # create test linops and vectors
    mat1 = torch.tensor([[1]], dtype=torch.float64)
    mat2 = torch.tensor([[1, 2], [3, 4]], dtype=torch.float64)
    mat3 = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=torch.float64)
    v1 = torch.ones_like(mat1[0])
    v2 = torch.ones_like(mat2[0])
    v3 = torch.ones_like(mat3[0])

    # 1x1 lower with diagonal
    tri = TriangularLinOp(mat1, 1, 1, lower=True, with_main_diagonal=True)
    assert torch.allclose(
        tri @ v1, torch.DoubleTensor((1.0,))
    ), "Incorrect result!"
    assert torch.allclose(
        v1 @ tri, torch.DoubleTensor((1.0,))
    ), "Incorrect result!"
    # 1x1 lower without diagonal (Hutch)
    tri = TriangularLinOp(mat1, 1, 1, lower=True, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v1, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"
    assert torch.allclose(
        v1 @ tri, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"
    # 1x1 lower without diagonal (Step)
    tri = TriangularLinOp(mat1, 1, 0, lower=True, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v1, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"
    assert torch.allclose(
        v1 @ tri, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"

    # 1x1 upper with diagonal
    tri = TriangularLinOp(mat1, 1, 1, lower=False, with_main_diagonal=True)
    assert torch.allclose(
        tri @ v1, torch.DoubleTensor((1.0,))
    ), "Incorrect result!"
    assert torch.allclose(
        v1 @ tri, torch.DoubleTensor((1.0,))
    ), "Incorrect result!"
    # 1x1 upper without diagonal (Hutch)
    tri = TriangularLinOp(mat1, 1, 1, lower=False, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v1, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"
    assert torch.allclose(
        v1 @ tri, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"
    # 1x1 upper without diagonal (Step)
    tri = TriangularLinOp(mat1, 1, 0, lower=False, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v1, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"
    assert torch.allclose(
        v1 @ tri, torch.DoubleTensor((0.0,))
    ), "Incorrect result!"

    # 2x2 lower with diagonal
    tri = TriangularLinOp(mat2, 2, 2, lower=True, with_main_diagonal=True)
    assert torch.allclose(
        tri @ v2, torch.DoubleTensor((1.0, 7))
    ), "Incorrect result!"
    assert torch.allclose(
        v2 @ tri, torch.DoubleTensor((4.0, 4))
    ), "Incorrect result!"
    # 2x2 lower without diagonal (Hutch)
    tri = TriangularLinOp(mat2, 2, 2, lower=True, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v2, torch.DoubleTensor((0.0, 3))
    ), "Incorrect result!"
    assert torch.allclose(
        v2 @ tri, torch.DoubleTensor((3.0, 0))
    ), "Incorrect result!"
    # 2x2 lower without diagonal (Step)
    tri = TriangularLinOp(mat2, 1, 0, lower=True, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v2, torch.DoubleTensor((0.0, 3))
    ), "Incorrect result!"
    assert torch.allclose(
        v2 @ tri, torch.DoubleTensor((3.0, 0))
    ), "Incorrect result!"

    # 2x2 upper with diagonal
    tri = TriangularLinOp(mat2, 2, 2, lower=False, with_main_diagonal=True)
    assert torch.allclose(
        tri @ v2, torch.DoubleTensor((3.0, 4))
    ), "Incorrect result!"
    assert torch.allclose(
        v2 @ tri, torch.DoubleTensor((1.0, 6))
    ), "Incorrect result!"
    # 2x2 upper without diagonal (Hutch)
    tri = TriangularLinOp(mat2, 2, 2, lower=False, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v2, torch.DoubleTensor((2.0, 0))
    ), "Incorrect result!"
    assert torch.allclose(
        v2 @ tri, torch.DoubleTensor((0.0, 2))
    ), "Incorrect result!"
    # 2x2 upper without diagonal (Step)
    tri = TriangularLinOp(mat2, 1, 0, lower=False, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v2, torch.DoubleTensor((2.0, 0))
    ), "Incorrect result!"
    assert torch.allclose(
        v2 @ tri, torch.DoubleTensor((0.0, 2))
    ), "Incorrect result!"

    # 3x3 lower with diagonal
    tri = TriangularLinOp(mat3, 3, 3, lower=True, with_main_diagonal=True)
    assert torch.allclose(
        tri @ v3, torch.DoubleTensor((1.0, 9, 24))
    ), "Incorrect result!"
    assert torch.allclose(
        v3 @ tri, torch.DoubleTensor((12.0, 13, 9))
    ), "Incorrect result!"
    # 3x3 lower without diagonal (Hutch)
    tri = TriangularLinOp(mat3, 3, 3, lower=True, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v3, torch.DoubleTensor((0.0, 4, 15))
    ), "Incorrect result!"
    assert torch.allclose(
        v3 @ tri, torch.DoubleTensor((11.0, 8, 0))
    ), "Incorrect result!"
    # 3x3 lower without diagonal (Step)
    tri = TriangularLinOp(mat3, 1, 0, lower=True, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v3, torch.DoubleTensor((0.0, 4, 15))
    ), "Incorrect result!"
    assert torch.allclose(
        v3 @ tri, torch.DoubleTensor((11.0, 8, 0))
    ), "Incorrect result!"

    # 3x3 upper with diagonal
    tri = TriangularLinOp(mat3, 3, 3, lower=False, with_main_diagonal=True)
    assert torch.allclose(
        tri @ v3, torch.DoubleTensor((6.0, 11, 9))
    ), "Incorrect result!"
    assert torch.allclose(
        v3 @ tri, torch.DoubleTensor((1.0, 7, 18))
    ), "Incorrect result!"
    # 3x3 upper without diagonal (Hutch)
    tri = TriangularLinOp(mat3, 3, 3, lower=False, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v3, torch.DoubleTensor((5.0, 6, 0))
    ), "Incorrect result!"
    assert torch.allclose(
        v3 @ tri, torch.DoubleTensor((0.0, 2, 9))
    ), "Incorrect result!"
    # 3x3 upper without diagonal (Step)
    tri = TriangularLinOp(mat3, 1, 0, lower=False, with_main_diagonal=False)
    assert torch.allclose(
        tri @ v3, torch.DoubleTensor((5.0, 6, 0))
    ), "Incorrect result!"
    assert torch.allclose(
        v3 @ tri, torch.DoubleTensor((0.0, 2, 9))
    ), "Incorrect result!"


def test_triangular_linop_corner_cases(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    dim_rank_decay_sym_width_hutch_maindiag_rtol,
    num_tri_tests,
):
    """Testing corner cases for triangular linop.

    Cases:

    * Setting step width to 0 raises an error
    * Performing 0 measurements is accepted, but yields zeros
    * Requesting too many measurements raises an error
    """
    order = 10
    mat = torch.randn((order, order), dtype=torch.float64)
    v = torch.ones_like(mat[0])
    z = torch.zeros_like(mat[0])
    for lower in (True, False):
        for diag in (True, False):
            # zero step width raises error
            with pytest.raises(AssertionError):
                TriangularLinOp(
                    mat,
                    0,
                    0,
                    lower=lower,
                    with_main_diagonal=diag,
                )
            # 9 stair measurements + 2 Hutch = 11 measurements. Error!
            with pytest.raises(AssertionError):
                TriangularLinOp(
                    mat,
                    1,
                    2,
                    lower=lower,
                    with_main_diagonal=diag,
                )
            # 0 stair measurements + 11 Hutch, also error
            with pytest.raises(AssertionError):
                TriangularLinOp(
                    mat,
                    order,
                    order + 1,
                    lower=lower,
                    with_main_diagonal=diag,
                )
            # no measurements yields zero operator
            tri = TriangularLinOp(
                mat,
                order,
                0,
                lower=lower,
                with_main_diagonal=diag,
            )
            assert torch.allclose(tri @ v, z), "Result should be zeros!"
            assert torch.allclose(v @ tri, z), "Result should be zeros!"


# ##############################################################################
# # SERRATED HADAMARD
# ##############################################################################
def test_serrated_hadamard_pattern(
    rng_seeds,  # noqa: F811
    hadamard_atol,
    torch_devices,  # noqa: F811
    serrated_hadamard_dims_chunks,
):
    """Test case for serrated Hadamard pattern generator.

    This particular Hadamard pattern is useful to estimate block-triangulars
    using Hutchinson.

    This test case checks that serrated patterns of vectors of ones match
    expected values for:
    * Lower and upper-triangular patterns
    * FFT and plain implementations
    * With and without diagonal
    * Different block sizes

    Then, it checks that FFT and plain implementations yield same results for
    random vectors.
    """
    for seed in rng_seeds:
        for dtype, atol in hadamard_atol.items():
            for device in torch_devices:
                # Simple tests with vector of ones:
                v = torch.ones(10, dtype=dtype, device=device)
                # chunk size must be 1 or greater
                with pytest.raises(ValueError):
                    serrated_hadamard_pattern(v, 0)
                #
                had = serrated_hadamard_pattern(v, 1, use_fft=False)
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(v, 2, use_fft=False)
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(v, 4, use_fft=False)
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [1, 2, 3, 4, 1, 2, 3, 4, 1, 2],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 1, with_main_diagonal=False, use_fft=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 2, with_main_diagonal=False, use_fft=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 4, with_main_diagonal=False, use_fft=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [0, 1, 2, 3, 0, 1, 2, 3, 0, 1],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 1, use_fft=False, lower=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 2, use_fft=False, lower=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [2, 1, 2, 1, 2, 1, 2, 1, 2, 1],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"

                had = serrated_hadamard_pattern(
                    v, 4, use_fft=False, lower=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [2, 1, 4, 3, 2, 1, 4, 3, 2, 1],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 1, with_main_diagonal=False, use_fft=False, lower=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 2, with_main_diagonal=False, use_fft=False, lower=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [1, 0, 1, 0, 1, 0, 1, 0, 1, 0],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                #
                had = serrated_hadamard_pattern(
                    v, 4, with_main_diagonal=False, use_fft=False, lower=False
                )
                assert torch.allclose(
                    had,
                    torch.tensor(
                        [1, 0, 3, 2, 1, 0, 3, 2, 1, 0],
                        dtype=dtype,
                        device=device,
                    ),
                    atol=atol,
                ), "Incorrect serrated result!"
                # random test with more complex responses
                for dims, chunk in serrated_hadamard_dims_chunks:
                    v = gaussian_noise(
                        dims,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                    )
                    for diag in (True, False):
                        for lower in (True, False):
                            had = serrated_hadamard_pattern(
                                v,
                                chunk,
                                with_main_diagonal=diag,
                                use_fft=False,
                                lower=lower,
                            )
                            had_fft = serrated_hadamard_pattern(
                                v,
                                chunk,
                                with_main_diagonal=diag,
                                use_fft=True,
                                lower=lower,
                            )
                            assert torch.allclose(
                                had, had_fft, atol=atol
                            ), "Inconsistent FFT implementation of serrated!"



# -*- coding: utf-8 -*-

r"""Distributed Sketched Decompositions
=======================================

In some cases, the main bottleneck to work with a linear operator :math:`A`
is performing linear measurements in the form :math:`w := A v`.

Luckily for us, sketched methods rely on multiple independent vectors
:math:`v_i`, which means that we can parallelize the measurements.

Another related bottleneck is that multiple processes may not be able to write
their  :math:`w_i` measurements to the same file concurrently, and if they
write to separate files, subsequent processing may require to merge them.

This tutorial showcases ``skerch`` functionality to overcome both these
problems and compute a distributed sketched SVD (DSSVD).

.. note::

  Readers are encouraged to read the :ref:`Sketched Decompositions` tutorial
  first and grasp the fundamental quantities and steps involved.
  This tutorial is a follow-up since it builds on top and naturally involves
  more implementation and algorithmic details, due to the distributed and
  system-agnostic nature of the interfaces.


Of course, all of the ``skerch`` functionality introduced in the other tutorials
can be applied to the distributed scenario, including *a priori* and
*a posteriori* methods. But to keep things simple, here we will:

* Compute the exact and sketched SVD of an asymmetric matrix
* Check that Frobenius error between original and recovery is low
* Check that :math:`\ell_2` error between original and recovered singular values
  is low

Let's start with slightly different imports: instead of directly importing
:func:`skerch.decompositions.ssvd`, we import its components:
"""


import tempfile

import matplotlib.pyplot as plt
import torch

from skerch.decompositions import truncate_core
from skerch.distributed_decompositions import (
    DistributedHDF5,
    create_hdf5_layout,
    orthogonalize,
    solve_core_ssvd,
)
from skerch.distributed_measurements import innermeas_idx_torch, ssrft_idx_torch
from skerch.ssrft import SSRFT
from skerch.synthmat import SynthMat

# %%
#
# ##############################################################################
#
# As before, we create and decompose a random test matrix.

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DTYPE = torch.float64
SHAPE, RANK = (200, 222), 10
OUTER, INNER, TRUNCATE = 30, 60, 20
NUM_A_POSTERIORI = 50
EXP_DECAY = 0.2
PROCESSING_FLAG, SUCCESS_FLAG = "0", "1"
SEED = 12345

lop = SynthMat.exp_decay(
    SHAPE,
    RANK,
    EXP_DECAY,
    symmetric=False,
    device=DEVICE,
    dtype=DTYPE,
    seed=SEED,
)
U_lop, S_lop, Vt_lop = torch.linalg.svd(lop)


# %%
#
# ##############################################################################
#
# In a distributed system, we assume that each measurement is performed and
# stored on a separate unit. This is handled by creating a virtual HDF5
# dataset (see :class:`skerch.distributed_measurements.DistributedHDF5` for
# more details).

tmpdir = tempfile.TemporaryDirectory()
(
    (lo_path, lo_subpaths),
    (ro_path, ro_subpaths),
    (inn_path, inn_subpaths),
) = create_hdf5_layout(
    tmpdir.name,
    SHAPE,
    lop.dtype,
    OUTER,
    INNER,
    lo_fmt="leftouter_{}.h5",
    ro_fmt="rightouter_{}.h5",
    inner_fmt="core_{}.h5",
    with_ro=True,
)


# %%
#
# ##############################################################################
#
# Now that we have our HDF5 layout in place, we can perform the distributed
# measurements. Although we have ``for`` loops here, you can imagine that
# each of the iterations is running in a separate process or device.
# They will be able to write concurrently to the distributed HDF5 layout.
#
# The loop over :func:`skerch.distributed_measurements.ssrft_idx_torch` performs
# a matrix-free random SSRFT projection :math:`A \Phi`, where :math:`A` is our
# original linear operator and each iteration is a column of the random
# projection :math:`\Phi` (see :mod:`skerch.ssrft` for more details).
#
# The :func:`skerch.distributed_measurements.innermeas_idx_torch` function is a
# slight variation in which each :math:`i^{th}` loop iteration computes
# :math:`\Psi^* A \Phi_i`, where :math:`\Psi` and :math:`\Phi` are SSRFT
# projections with different seeds (i.e. uncorrelated to each other).

# SSRFT random seeds
lo_seed, ro_seed, li_seed, ri_seed = SEED + 1, SEED + 2, SEED + 3, SEED + 4

# left outer measurements
for i in range(OUTER):
    vals, flag, h5 = DistributedHDF5.load(lo_subpaths[i])
    ssrft_idx_torch(
        i,
        OUTER,
        lop,
        DEVICE,
        DTYPE,
        vals,
        lo_seed,
        flag=flag,
        processing_flag=PROCESSING_FLAG,
        success_flag=SUCCESS_FLAG,
        adjoint=True,
    )
    # check that file has been successfully written to
    assert flag[0].decode() == SUCCESS_FLAG, f"Unsuccessful measurement? {i}"
    h5.close()

# right outer measurements
for i in range(OUTER):
    vals, flag, h5 = DistributedHDF5.load(ro_subpaths[i])
    ssrft_idx_torch(
        i,
        OUTER,
        lop,
        DEVICE,
        DTYPE,
        vals,
        ro_seed,
        flag=flag,
        processing_flag=PROCESSING_FLAG,
        success_flag=SUCCESS_FLAG,
        adjoint=False,
    )
    # check that file has been successfully written to
    assert flag[0].decode() == SUCCESS_FLAG, f"Unsuccessful measurement? {i}"
    h5.close()

# inner measurements
for i in range(INNER):
    vals, flag, h5 = DistributedHDF5.load(inn_subpaths[i])
    innermeas_idx_torch(
        i,
        INNER,
        lop,
        DEVICE,
        DTYPE,
        vals,
        li_seed,
        ri_seed,
        flag=flag,
        processing_flag=PROCESSING_FLAG,
        success_flag=SUCCESS_FLAG,
    )
    # check that file has been successfully written to
    assert flag[0].decode() == SUCCESS_FLAG, f"Unsuccessful measurement? {i}"
    h5.close()


# %%
#
# Once we are done with the measurements, merge the respective HDF5 files into
# a centralized one. This is to avoid OS errors (see
# :class:`skerch.distributed_measurements.DistributedHDF5` for an
# explanation).
#
# Since we are deleting each measurement after merging it, there is no
# substantial memory overhead and we simply "repack" existing data.

_ = DistributedHDF5.merge_all(
    lo_path,
    out_path=None,
    delete_subfiles_while_merging=True,
    check_success_flag=SUCCESS_FLAG,
)
_ = DistributedHDF5.merge_all(
    ro_path,
    out_path=None,
    delete_subfiles_while_merging=True,
    check_success_flag=SUCCESS_FLAG,
)
_ = DistributedHDF5.merge_all(
    inn_path,
    out_path=None,
    delete_subfiles_while_merging=True,
    check_success_flag=SUCCESS_FLAG,
)


# %%
#
# ##############################################################################
#
# The most conceptually involved step in the sketched SVD is to "solve the core
# matrix". This is done in a centralized manner by taking the QR decomposition
# of the outer measurements and solving two least squares problems against
# the inner measurements (see `[TYUC2019] <https://arxiv.org/abs/1902.08651>`_).
#
# To speed things up, and in the lack of an out-of-core QR routine, this needs
# to be done on a single computer with sufficient RAM to perform the QR step.

# load all measurements and check that all flags are OK
lo, lo_flags, lo_h5 = DistributedHDF5.load(lo_path, filemode="r+")
ro, ro_flags, ro_h5 = DistributedHDF5.load(ro_path, filemode="r+")
inn, inn_flags, inn_h5 = DistributedHDF5.load(inn_path, filemode="r+")
assert all(
    f.decode() == SUCCESS_FLAG for f in lo_flags
), "Bad left outer flags!"
assert all(
    f.decode() == SUCCESS_FLAG for f in ro_flags
), "Bad right outer flags!"
assert all(f.decode() == SUCCESS_FLAG for f in inn_flags), "Bad inner flags!"

# orthogonalize outer measurements (in-place)
orthogonalize(lo, overwrite=True)
orthogonalize(ro, overwrite=True)

# Solve core op and decompose
left_inner_ssrft = SSRFT((INNER, SHAPE[0]), seed=li_seed)
right_inner_ssrft = SSRFT((INNER, SHAPE[1]), seed=ri_seed)
core_U, core_S, core_Vt = solve_core_ssvd(
    ro, lo, inn, left_inner_ssrft, right_inner_ssrft
)
# for this utest, convert everything to explicit torch tensors
ro = torch.from_numpy(ro[:]).to(DEVICE)
lo = torch.from_numpy(lo[:]).to(DEVICE)
core_U = torch.from_numpy(core_U).to(DEVICE)
core_S = torch.from_numpy(core_S).to(DEVICE)
core_Vt = torch.from_numpy(core_Vt).to(DEVICE)
core_U, core_S, core_Vt = truncate_core(TRUNCATE, core_U, core_S, core_Vt)
# close HDF5 files and we are done
lo_h5.close()
ro_h5.close()
inn_h5.close()


# %%
#
# ##############################################################################
#
# At this point we are done with the sketched SVD! To naively test the quality
# of the recovered decomposition, we measure the error overall and for the
# singular values, verifying that they are low:

appr_lop = ro @ (core_U @ torch.diag(core_S) @ core_Vt) @ lo.T

# plot part of original and recovery
fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(lop.cpu()[:50, :50], cmap="bwr")
ax2.imshow(appr_lop.cpu()[:50, :50], cmap="bwr")
fig.tight_layout()

print("Frobenius error:", torch.dist(lop, appr_lop).item())
print(
    "Singular value error:",
    torch.dist(S_lop[:TRUNCATE], core_S[:TRUNCATE]).item(),
)

# Done!
tmpdir.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Command Line Interface interaction with ``skerch``.

This script acts in connection with the :mod:`.cli` submodule.
"""


import argparse
import pkgutil
import shutil
import sys
import tempfile
from unittest.mock import patch

from . import INNER_FMT, LO_FMT, RO_FMT, cli


# ##############################################################################
# # HELPERS
# ##############################################################################
def ensure_dirpath(dirpath=None):
    """Produces a valid directory path if none given.

    If ``dirpath`` is not None, returns ``dirpath``.
    If ``dirpath`` is None, prompts user to accept creating a random temporary
    directory. If user accepts directory is created and its path returned,
    otherwise process is exited.
    """
    if dirpath is None:
        dirpath = tempfile.mkdtemp()
        ans = input(f"No dirpath provided! proceed with {dirpath}? (y/N): ")
        if ans.lower() != "y":
            shutil.rmtree(dirpath)
            sys.exit(0)
    return dirpath


class PluginLoader:
    """Plugin helper that loads functionality from :cvar:`PLUGIN_MODULE`."""

    PLUGIN_MODULE = cli
    ENTRY_POINT = "main"

    @classmethod
    def get(cls):
        """Loads all plugins from :mod:`.cli`.

        :returns: A dictionary in the form ``plugin_name: plugin_fn``, where
          each name corresponds to a file in the :mod:`.cli` submodule, and
          the corresponding function is the :cvar:`ENTRY_POINT` method in that
          file.
        """
        result = {}
        modname = cls.PLUGIN_MODULE.__name__
        for sub in pkgutil.iter_modules(cls.PLUGIN_MODULE.__path__):
            subname = sub.name
            # ridiculously obscure line to allow submodules from __main__,
            # when they make relative imports themselves.
            submod = __import__(
                f"{modname}.{subname}",
                globals(),
                locals(),
                # level=1,
                fromlist=[""],
            )
            result[subname] = getattr(submod, cls.ENTRY_POINT)
        #
        return result


# ##############################################################################
# # GLOBALS
# ##############################################################################
COMMANDS = PluginLoader.get()  # parse the .cli module and retrieve functions


# ##############################################################################
# # PARSE CLI
# ##############################################################################
def matrix_shape(shape):
    """Checks and parses given string into a shape.

    If given ``shape`` is a string with two positive, coma-separated integers
    like '100,200', returns a tuple in the form ``(100, 200)``. Otherwise
    raises an ``AssertionError``.
    """
    h, w = shape.split(",")
    h, w = int(h), int(w)
    assert h > 0, "Height must be positive!"
    assert w > 0, "Width must be positive!"
    return (h, w)


def get_argparser():
    """Argument parser for ``skerch`` CLI interaction.

    :returns: An ``ArgumentParser`` that first accepts a positional argument
      with the ``command`` to be executed (must be one of the dict keys
      returned by :meth:`PluginLoader.get`), and then accepts optional
      arguments that may be applicable to that particular command. Call
      with ``-h`` for details.

    Usage examples::
      python -m skerch post_bounds --apost_n=30 --apost_err=0.75
      python -m skerch prio_hpars --shape=100,200 --budget=12345
      python -m skerch create_hdf5_layout --dtype float64 \
        --shape 123,234 --outer 30 --inner 60
    """
    parser = argparse.ArgumentParser(description="skerch CLI")
    parser.add_argument(
        "command",
        metavar="[COMMAND]",
        action="store",
        type=str,
        choices=set(COMMANDS),
        help=f"Determines which functionality to run: {set(COMMANDS)}",
    )
    # a priori parameter estimation
    parser.add_argument(
        "--shape",
        type=matrix_shape,
        default=(100, 200),
        help="Matrix shape in the form 'height,width' as positive integers.",
    )
    parser.add_argument(
        "--budget",
        type=int,
        default=10_000,
        help="Total number of scalar measurements that we can store.",
    )
    parser.add_argument(
        "--sym",
        action="store_true",  # default is false
        help="If given, matrix is assumed to be Hermitian.",
    )
    # a posteriori error bounds
    parser.add_argument(
        "--apost_n",
        default=30,
        type=int,
        help="Number of a-posteriori measurements intended.",
    )
    parser.add_argument(
        "--apost_err",
        default=0.5,
        type=float,
        help="A-posteriori target error, from 0 (no error) to 1 (0x - 2x).",
    )
    # create HDF5 layout
    parser.add_argument(
        "--hdf5dir",
        default=None,
        type=str,
        help="Directory to create the HDF5 layout.",
    )
    parser.add_argument(
        "--outer", default=100, type=int, help="Number of outer measurements."
    )
    parser.add_argument(
        "--inner", default=200, type=int, help="Number of inner measurements."
    )
    parser.add_argument(
        "--dtype",
        default="float32",
        type=str,
        help="Datatype of HDF5 layout to be created. Must be empty",
    )
    # merge virtual HDF5 into monolithic
    parser.add_argument(
        "--in_path",
        default=None,
        type=str,
        help="Input path for the file to be processed.",
    )
    parser.add_argument(
        "--out_path",
        default=None,
        type=str,
        help="Input path for the file to be processed.",
    )
    parser.add_argument(
        "--ok_flag",
        default=None,
        type=str,
        help="If given, all HDF5 flags are checked to equal this.",
    )
    parser.add_argument(
        "--delete_subfiles",
        action="store_true",
        help="If given, HDF5 subfiles are deleted upon merging to monolithic.",
    )
    return parser


# ##############################################################################
# # MAIN ROUTINE
# ##############################################################################
def main_wrapper(cli_args=None):
    """Wrapped :func:`main` function.

    This wrapper gives the option to programmatically mock CLI arguments if
    desired. When running normally, e.g. via ``python -m ...``, it does
    nothing.

    :param cli_args: If none given, CLI arguments from ``sys.argv`` are left
      untouched. Otherwise, ``sys.argv[1:]`` will be mocked to have the given
      values.

    The following snippet has the same effect as running
    ``python -m skerch post_bounds --apost_n=30 --apost_err=0.75`` from the
    CLI::

      from skerch.__main__ import main_wrapper as skm
      skm(["post_bounds", "--apost_n=30", "--apost_err=0.75"])
    """
    if cli_args is None:
        main()
    else:
        cli_args = [sys.argv[0]] + list(cli_args)
        with patch.object(sys, "argv", cli_args):
            try:
                main()
            except SystemExit:
                # ArgParser exits with '-h', we don't want that
                # https://stackoverflow.com/a/58367457
                pass


def main():
    """Entry point for ``python -m skerch``."""
    # parse args and check that action is recognized
    parser = get_argparser()
    args = parser.parse_args()
    cmd = args.command
    assert cmd in COMMANDS, f"Unknown command! {cmd}"
    main = COMMANDS[cmd]
    # a-priori estimation of hyperparameters
    if cmd == "prio_hpars":
        shape = args.shape
        budget = args.budget
        sym = args.sym
        main(shape, budget, sym)
    # a-posteriori error probability bounds
    elif cmd == "post_bounds":
        n = args.apost_n
        err = args.apost_err
        main(n, err)
    # create HDF5 layout
    elif cmd == "create_hdf5_layout":
        dirpath = ensure_dirpath(args.hdf5dir)
        shape = args.shape
        with_ro = not args.sym  # if matrix Hermitian, no need for right outer
        outer = args.outer
        inner = args.inner
        dtype = args.dtype
        main(
            dirpath,
            shape,
            dtype,
            outer,
            inner,
            LO_FMT,
            RO_FMT,
            INNER_FMT,
            with_ro,
        )
    # merge
    elif cmd == "merge_hdf5":
        inpath = args.in_path
        outpath = args.out_path
        flag = args.ok_flag
        delete_sub = args.delete_subfiles
        main(inpath, outpath, flag, delete_sub)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""A-priori utilities for sketched decompositions.

Given a limited amount of memory, we would like to know how many inner and
outer measurements we should perform to get the best results.

In `[TYUC2019, 5.4.2] <https://arxiv.org/abs/1902.08651>`_, it is argued that
the number of outer measurements should be maximized within memory constraints,
subject to the number of inner measurements being twice.
This yields a closed-form optimum for general matrices, presented in the paper
and implemented here.
"""


import math


# ##############################################################################
# # A PRIORI HYPERPARAMETERS
# ##############################################################################
def a_priori_hyperparams(
    matrix_shape, memory_budget, complex_data=False, hermitian=False
):
    """A-priori hyperparameter selection for sketched measurements.

    Given the shape of a generic linear operator that we wish to decompose, and
    a memory budget, return the number of inner and outer measurements required,
    such that the budget is not exceeded, ``inner >= 2 * outer``, and outer is
    maximized (see `[TYUC2019, 5.4.2] <https://arxiv.org/abs/1902.08651>`_).

    :param int memory_budget: Memory available in number of matrix entries.
      E.g. if the matrix is in float32, a budget of ``N`` means ``4N`` bytes.
    :param hermitian: If true, the matrix is assumed to be hermitian (and
      square), so measurements only need to be done on one side. This means
      that, for the same budget, we can do more measurements.
    :returns: The pair ``(k, s)``, where the first integer is the optimal
      number of outer sketch measurements, and the second one is the
      corresponding number of core measurements. Optimality is defined as:
      distribute the budget such that k is maximized, but ``s >= 2k``
      (see 5.4.2. in paper).
    """
    m, n = matrix_shape
    if hermitian:
        assert m == n, "Hermitian matrix must be also square!"
        n = 0  # if matrix is hermitian ,we only count m, not n
    alpha = 0 if complex_data else 1
    mn4a = m + n + 4 * alpha
    budget_root = 16 * (memory_budget - alpha**2)
    #
    outer_dim = math.floor((1 / 8) * (math.sqrt(mn4a**2 + budget_root) - mn4a))
    core_dim = math.floor(math.sqrt(memory_budget - outer_dim * (m + n)))
    #
    return outer_dim, core_dim



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Functionality to support sketched estimation of any (sub-)diagonals.

Including:

* Hadamard pattern to compute subdiagonals as ``mean(hadamard(v) * lop @ v)``,
  for ``v_i = SSRFT[i, :]``.
* In-core routine to compute arbitrary subdiagonals of linops.
"""


import torch

from .distributed_decompositions import orthogonalize
from .linops import CompositeLinOp, NegOrthProjLinOp
from .ssrft import SSRFT


# ##############################################################################
# # HADAMARD
# ##############################################################################
def subdiag_hadamard_pattern(v, diag_idxs, use_fft=False):
    r"""Shifted copies of vectors for subdiagonal Hutchinson estimation.

    Given a square linear operator :math:`A`, and random vectors
    :math:`v \sim \mathcal{R}` with :math:`\mathbb{E}[v v^T] = I`, consider
    this generalized formulation of the Hutchinson diagonal estimator:

    .. math::

      f(A) =
      \mathbb{E}_{v \sim \mathcal{R}} \big[ \varphi(v) \odot Av \big]

    If the :math:`\varphi` function is the identity, then :math:`f(A)` equals
    the main diagonal of :math:`A`. If e.g. :math:`\varphi` shifts the entries
    in :math:`v` downwards by ``k`` positions, we get the ``k``-th subdiagonal.

    .. seealso::

      `[BN2022] <https://arxiv.org/abs/2201.10684>`_: Robert A. Baston and Yuji
      Nakatsukasa. 2022. *“Stochastic diagonal estimation: probabilistic bounds
      and an improved algorithm”*.  CoRR abs/2201.10684.

    :param v: Torch vector expected to contain zero-mean, uncorrelated entries.
    :param diag_idxs: Iterator with integers corresponding to the subdiagonal
      indices to include, e.g. 0 corresponds to the main diagonal, 1 to the
      diagonal above, -1 to the diagonal below, and so on.
    :param use_fft: If false, shifted copies of ``v`` are pasted on the result.
      This requires only ``len(v)``  memory, but has ``len(v) * len(diag_idxs)``
      time complexity. If this argument is true, an FFT convolution is used
      instead. This requires at least ``4 * len(v)`` memory, but the arithmetic
      has a complexity of ``len(v) * log(len(v))``, which can be advantageous
      whenever ``len(diag_idxs)`` becomes very large.
    :returns: A vector of same shape, type and device as ``v``, composed of
      shifted copies of ``v`` as given by ``diag_idxs``.
    """
    len_v = len(v)
    if use_fft:
        # create a buffer of zeros to avoid circular conv and store the
        # convolutional impulse response
        buff = torch.zeros(2 * len_v, dtype=v.dtype, device=v.device)
        # padded FFT to avoid circular convolution
        buff[:len_v] = v
        V = torch.fft.rfft(buff)
        # now we can write the impulse response on buff
        buff[:len_v] = 0
        for idx in diag_idxs:
            buff[idx] = 1
        # non-circular FFT convolution:
        V *= torch.fft.rfft(buff)
        V = torch.fft.irfft(V)[:len_v]
        return V
    else:
        result = torch.zeros_like(v)
        for idx in diag_idxs:
            if idx == 0:
                result += v
            elif idx > 0:
                result[idx:] += v[:-idx]
            else:
                result[:idx] += v[-idx:]
        return result


# ##############################################################################
# # SKETCHED (SUB)-DIAGONAL ESTIMATOR
# ##############################################################################
def subdiagpp(
    lop,
    num_meas,
    lop_dtype,
    lop_device,
    seed=0b1110101001010101011,
    deflation_rank=0,
    diag_idx=0,
):
    """Matrix-free (sub-) diagonal sketched estimator via Diag++.

    .. seealso::

      Reference for this algorithm:

      `[BN2022] <https://arxiv.org/abs/2201.10684>`_: Robert A. Baston and Yuji
      Nakatsukasa. 2022. *“Stochastic diagonal estimation: probabilistic bounds
      and an improved algorithm”*.  CoRR abs/2201.10684.

    Given a linear operator ``lop``, implements an unbiased diagonal estimator,
    composed of orthogonal deflation followed by Hutchinson estimation.

    The (optional) deflation is based on efficiently obtaining a thin matrix of
    orthonormal columns ``Q``, such that:

    ``lop = (Q @ Q.T @ lop) + ((I - Q@Q.T) @ lop)``

    Then, the (sub-) diagonal entries for the first component can be computed
    exactly and efficiently by picking rows of ``Q``. The second, "deflated"
    term is then subject to Hutchinson estimation:

    ``diag_k(defl_lop) = expectation(hadamard_k(v) * defl_lop @ v)``,

    This is achieved by sampling random ``v`` vectors from a SSRFT operator,
    and shifting them using :func:`subdiag_hadamard_pattern` to capture the
    desired ``k``-th subdiagonal.

    The final estimation is then the sum of the exact first component plus the
    estimated, deflated component.

    :param lop: The linear operator to extract (sub-) diagonals from. It must
      implement a ``lop.shape = (h, w)`` attribute as well as the left- and
      right- matmul operator ``@``, interacting with torch tensors.
    :param num_meas: Number of samples ``v`` used to compute the expectation.
    :param lop_dtype: Datatype of ``lop``.
    :param lop_device: Device of ``lop``.
    :param seed: Random seed for the ``v`` random SSRFT vectors.
    :param deflation_rank: Rank of the deflation matrix ``Q`` to be computed.
    :param diag_idx: Position of the (sub-) diagonal to compute. 0 corresponds
      to the main diagonal, +1 to the diagonal above, -1 to the diagonal below,
      and so on. Note that if diagonals are too far away from the main diagonal
      (or if ``lop`` is small enough), it may be worth it to directly perform
      one exact measurement per diagonal entry and not use ``Diag++`` at all.
    :returns: The tuple ``(result, defl, (top_norm, bottom_norm))``. The first
      element is the estimated diagonal. The second is the computed deflation
      matrix, or ``None`` if ``deflation_rank=0``. The last element is the
      Euclidean norm of the deflation and the deflated parts of the diagonal,
      respectively, which can be used to roughly diagnose the contribution of
      each part of the algorithm.
    """
    h, w = lop.shape
    assert h == w, "Only square linear operators supported!"
    dims = h
    abs_diag_idx = abs(diag_idx)
    result_buff = torch.zeros(
        dims - abs_diag_idx, dtype=lop_dtype, device=lop_device
    )
    deflation_matrix = None
    # first compute top-rank orth proj to deflate lop
    if deflation_rank > 0:
        # deflate lop: first compute a few random measurements
        ssrft_defl = SSRFT((deflation_rank, dims), seed=seed + 100)
        deflation_matrix = torch.empty(
            (dims, deflation_rank), dtype=lop_dtype, device=lop_device
        )
        for i in range(deflation_rank):
            deflation_matrix[:, i] = lop @ ssrft_defl.get_row(
                i, lop_dtype, lop_device
            )
        # orthogonalize measurements to get deflated lop
        orthogonalize(deflation_matrix, overwrite=True)
        negproj = NegOrthProjLinOp(deflation_matrix)
        deflated_lop = CompositeLinOp((("negproj", negproj), ("lop", lop)))
    else:
        # no deflation
        deflated_lop = lop
    # estimate deflated diagonal. We don't use subdiag_hadamard_pattern
    # because this direct way is more memory-efficient.
    if num_meas > 0:
        ssrft = SSRFT((num_meas, dims), seed=seed)
        for i in range(num_meas):
            v = ssrft.get_row(i, lop_dtype, lop_device)
            if diag_idx == 0:
                result_buff += v * (v @ deflated_lop)
                # squares_buff += v * v
            elif diag_idx > 0:
                result_buff += (
                    v[:-abs_diag_idx] * (v @ deflated_lop)[abs_diag_idx:]
                )
            elif diag_idx < 0:
                result_buff += (
                    v[abs_diag_idx:] * (v @ deflated_lop)[:-abs_diag_idx]
                )
        # note that here we would typically apply result_buff /= squares_buff
        # but l2 norm of SSRFT rows is always 1, so no need.
    bottom_norm = result_buff.norm().item()
    # add estimated deflated diagonal to exact top-rank diagonal
    top_norm = 0
    if deflation_rank > 0:
        for i in range(len(result_buff)):
            row = i if (diag_idx > 0) else i + abs_diag_idx
            col = i if (diag_idx <= 0) else i + abs_diag_idx
            entry = ((deflation_matrix @ deflation_matrix[row]) @ lop)[col]
            result_buff[i] += entry
            top_norm += entry**2
        top_norm = top_norm**0.5
    #
    return result_buff, deflation_matrix, (top_norm, bottom_norm)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""CLI plugin to create a HDF5 layout to hold outer and inner measurements."""


import os

import torch

from ..distributed_decompositions import create_hdf5_layout


# ##############################################################################
# # ENTRY POINT
# ##############################################################################
def main(
    dirpath, shape, lop_dtype, outer, inner, lo_fmt, ro_fmt, inner_fmt, with_ro
):
    """See :func:`.distributed_decompositions.create_hdf5_layout`."""
    h, w = shape
    if (h != w) and (not with_ro):
        raise AssertionError("Non-square matrix can't be symmetric!")
    assert not os.listdir(dirpath), "Directory must be empty!"
    #
    lop_dtype = getattr(torch, lop_dtype)
    (
        (lo_path, lo_subpaths),
        (ro_path, ro_subpaths),
        (inner_path, inner_subpaths),
    ) = create_hdf5_layout(
        dirpath,
        shape,
        lop_dtype,
        outer,
        inner,
        lo_fmt=lo_fmt,
        ro_fmt=ro_fmt,
        inner_fmt=inner_fmt,
        with_ro=with_ro,
    )
    print(dirpath)
    print(shape)
    print(lop_dtype)
    print(outer)
    print(inner)
    print(lo_fmt)
    print(ro_fmt)
    print(inner_fmt)
    print(with_ro)
    #
    print(lo_path)
    print(ro_path)
    print(inner_path)
    # print(lo_subpaths)
    # print(ro_subpaths)
    # print(inner_subpaths)



# -*- coding: utf-8 -*-

r"""Sketched Decompositions
===========================

In this tutorial we create noisy, numerically low-rank matrices,
and compare their exact decomposition with the sketched one.

We do this for arbitrary matrices (SVD), as well as symmetric
ones (Hermitian EIGH), both PSD and non-PSD.

We also show how to assess rank and quality of the recovery in more difficult
scenarios. For that, a-posteriori error estimation methods are used.
"""

import matplotlib.pyplot as plt
import torch

from skerch.a_posteriori import (
    a_posteriori_error,
    a_posteriori_error_bounds,
    scree_bounds,
)
from skerch.decompositions import seigh, ssvd, truncate_core
from skerch.linops import CompositeLinOp, DiagonalLinOp
from skerch.synthmat import SynthMat

# %%
#
# ##############################################################################
#
# Global variables and hyperparameters
# ------------------------------------
#
# ``OUTER`` and ``INNER`` reflect the number of sketches to take. Typically,
# ``INNER`` is twice ``OUTER``, which is itself larger than ``RANK`` by a
# constant (see `[TYUC2019] <https://arxiv.org/abs/1902.08651>`_).
#
# Despite the availability of *a priori* methods to estimate the number of
# measurements, (see :mod:`skerch.a_priori`), ``RANK`` is typically unknown
# and we resort to *a posteriori* methods, which estimate the quality of
# the approximation as a function of the recovered rank using
# ``NUM_A_POSTERIORI`` measurements.
#
# Last but not least, recovered singular vectors can get noisy when the
# associated singular values get closer to zero. Generally, this is tackled
# via truncation of the core matrices down to ``TRUNCATE`` dimensions.

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DTYPE = torch.float64
SHAPE, RANK = (200, 200), 10
OUTER, INNER, TRUNCATE = 30, 60, 20
NUM_A_POSTERIORI = 50
EXP_DECAY = 0.2

# %%
#
# ##############################################################################
#
# Creation of test data
# ---------------------
#
# Now we create three random, rank-defficient matrices, symmetric
# PSD and non-PSD (see :mod:`skerch.synthmat`). Then compute their exact
# decompositions using ``torch``:

linop_asym = SynthMat.exp_decay(
    SHAPE, RANK, EXP_DECAY, symmetric=False, device=DEVICE, dtype=DTYPE
)
linop_sym = SynthMat.exp_decay(
    SHAPE,
    RANK,
    EXP_DECAY,
    symmetric=True,
    device=DEVICE,
    dtype=DTYPE,
    psd=False,
)
linop_psd = SynthMat.exp_decay(
    SHAPE, RANK, EXP_DECAY, symmetric=True, device=DEVICE, dtype=DTYPE, psd=True
)

U_asym, S_asym, Vt_asym = torch.linalg.svd(linop_asym)
S_sym, U_sym = torch.linalg.eigh(linop_sym)
S_psd, U_psd = torch.linalg.eigh(linop_psd)

# %%
#
# ##############################################################################
#
# Computation of sketched approximations
# --------------------------------------
#
# We can now conveniently compute and truncate in-core sketched decompositions
# using functionality from :mod:`skerch.decompositions`. We can also efficiently
# compose the recovered thin matrices into the original shape (see documentation
# of :class:`skerch.linops.CompositeLinOp` for more details).

q_asym, u_asym, s_asym, vt_asym, pt_asym = ssvd(
    linop_asym,
    op_device=DEVICE,
    op_dtype=DTYPE,
    outer_dim=OUTER,
    inner_dim=INNER,
)
u_asym, s_asym, vt_asym = truncate_core(TRUNCATE, u_asym, s_asym, vt_asym)
appr_asym = CompositeLinOp(
    (
        ("Q", q_asym),
        ("U", u_asym),
        ("S", DiagonalLinOp(s_asym)),
        ("Vt", vt_asym),
        ("Pt", pt_asym),
    )
)

q_sym, u_sym, s_sym = seigh(
    linop_sym,
    op_device=DEVICE,
    op_dtype=DTYPE,
    outer_dim=OUTER,
    inner_dim=INNER,
)
u_sym, s_sym = truncate_core(TRUNCATE, u_sym, s_sym)
appr_sym = CompositeLinOp(
    (
        ("Q", q_sym),
        ("U", u_sym),
        ("S", DiagonalLinOp(s_sym)),
        ("Ut", u_sym.T),
        ("Qt", q_sym.T),
    )
)

q_psd, u_psd, s_psd = seigh(
    linop_psd,
    op_device=DEVICE,
    op_dtype=DTYPE,
    outer_dim=OUTER,
    inner_dim=INNER,
)
u_psd, s_psd = truncate_core(TRUNCATE, u_psd, s_psd)
appr_psd = CompositeLinOp(
    (
        ("Q", q_psd),
        ("U", u_psd),
        ("S", DiagonalLinOp(s_psd)),
        ("Ut", u_psd.T),
        ("Qt", q_psd.T),
    )
)

# %%
#
# ##############################################################################
#
# Exact error analysis
# -------------------->
#
# Since we are working with small matrices, we can here naïvely compare
# original matrix and reconstruction in terms of their Frobenius error, and
# even plot them:

naive_asym = q_asym @ u_asym @ DiagonalLinOp(s_asym) @ vt_asym @ pt_asym
naive_sym = q_sym @ u_sym @ DiagonalLinOp(s_sym) @ u_sym.T @ q_sym.T
naive_psd = q_psd @ u_psd @ DiagonalLinOp(s_psd) @ u_psd.T @ q_psd.T

fig, (ax1, ax2) = plt.subplots(ncols=2)
ax1.imshow(linop_asym.cpu()[:50, :50], cmap="bwr")
ax2.imshow(naive_asym.cpu()[:50, :50], cmap="bwr")
fig.tight_layout()

print("Frobenius error (asym):", torch.dist(linop_asym, naive_asym).item())
print("Frobenius error (sym):", torch.dist(linop_sym, naive_sym).item())
print("Frobenius error (psd):", torch.dist(linop_psd, naive_psd).item())
print(
    "Singular value error (asym):",
    torch.dist(S_asym[:TRUNCATE], s_asym[:TRUNCATE]).item(),
)
# Note that torch.eigh provides eigenpairs in ascending value, but we want
# them by descending magnitude
S_sym_descmag = S_sym[abs(S_sym).argsort(descending=True)]
print(
    "Eigenvalue error (sym):",
    torch.dist(S_sym_descmag[:TRUNCATE], s_sym[:TRUNCATE]).item(),
)
S_psd_descmag = S_psd[abs(S_psd).argsort(descending=True)]
print(
    "Eigenvalue error (psd):",
    torch.dist(S_psd_descmag[:TRUNCATE], s_psd[:TRUNCATE]).item(),
)

# %%
#
# ##############################################################################
#
# Approximate error analysis
# --------------------------
#
# When the linear operator is too large or matrix-free, an exact error analysis
# is not always possible, and we resort to approximate *a posteriori* methods
# (see :mod:`skerch.a_posteriori` for more details).
#
# First, we estimate the Frobenius error between the orginal and the sketched
# operators. Note how this does not require an explicit reconstruction.

(f1_asym, f2_asym, frob_err_asym) = a_posteriori_error(
    linop_asym, appr_asym, NUM_A_POSTERIORI, dtype=DTYPE, device=DEVICE
)[0]
print("Estimated Frobenius Error (asym):", frob_err_asym**0.5)

(f1_sym, f2_sym, frob_err_sym) = a_posteriori_error(
    linop_sym, appr_sym, NUM_A_POSTERIORI, dtype=DTYPE, device=DEVICE
)[0]
print("Estimated Frobenius Error (sym):", frob_err_sym**0.5)

(f1_psd, f2_psd, frob_err_psd) = a_posteriori_error(
    linop_psd, appr_psd, NUM_A_POSTERIORI, dtype=DTYPE, device=DEVICE
)[0]
print("Estimated Frobenius Error (psd):", frob_err_psd**0.5)


# %%
#
# We see that the approximate errors are very close to the previously computed
# exact ones. The probability of this not happening, for any given number
# of *a posteriori* measurements and error tolerance (in this case 50%), can
# be obtained as follows:

a_posteriori_error_bounds(NUM_A_POSTERIORI, 0.5)


# %%
#
# ##############################################################################
#
# *A posteriori* rank estimation
# ------------------------------
#
# With the quantities obtained so far, we can compute the *scree* bounds, which
# tell us how much of the Frobenius norm of the original operator are we
# actually recovering, as a function of increasing number of dimensions.
#
# This information can be used to estimate the effective rank of the original
# operator, e.g. by looking for "elbows" in the scree curve. Since in this
# case we have a simple synthetic matrix, we can see how the scree bounds
# accurately and clearly reflect the given ``RANK`` of the original operator.
#
# .. warning::
#
#   This functionality has not been unit-tested, use with care. Contributions
#   are welcome :)

scree_lo, scree_hi = scree_bounds(s_asym, f1_asym**0.5, frob_err_asym**0.5)
plt.plot(scree_lo.cpu())
plt.plot(scree_hi.cpu())


# %%
#
# ##############################################################################
#
# And we are done!
#
# * We have seen how to approximate singular (eigen-) decompositions via
#   in-core sketched methods for a broad class of low-rank matrices
# * We verified that said approximations are actually close to the original
#   ones
# * We demonstrated matrix-free, scalable methods to also verify the quality
#   of the approximation and the rank of the original operators



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Sketched decompositions on distributed systems.

For in-core decompositions, see :mod:`.distributed_decompositions`.
"""


import os

import numpy as np
import scipy
import torch

from .distributed_measurements import DistributedHDF5
from .linops import DiagonalLinOp

#
from .utils import torch_dtype_as_str


# ##############################################################################
# # CREATE HDF5 LAYOUTS
# ##############################################################################
def create_hdf5_layout(
    dirpath,
    lop_shape,
    lop_dtype,
    num_outer_measurements,
    num_inner_measurements,
    lo_fmt="leftouter_{}.h5",
    ro_fmt="rightouter_{}.h5",
    inner_fmt="inner_{}.h5",
    with_ro=True,
):
    """Creation of persistent HDF5 files to store sketches of a linear operator.

    :param str dirpath: Where to store the HDF5 files.
    :param lop_shape: Shape of linear operator to sketch from, in the form
      ``(height, width)``.
    :param lop_dtype: Torch dtype of the operator, e.g. ``torch.float32``. The
      HDF5 arrays will be of same type.
    :param int num_outer_measurements: Left outer measurement layout contains
      ``(width, outer)`` entries, and right outer layout ``(height, outer)``.
    :param int num_inner_measurements: Inner measurement layout contains
      ``(inner, inner)`` entries.
    :lo_fmt: Format string for the left-outer HDF5 filenames.
    :ro_fmt: Format string for the right-outer HDF5 filenames.
    :inner_fmt: Format string for the inner HDF5 filenames.
    :param with_ro: If false, no right outer layout will be created (useful
      when working with symmetric matrices where only one side is needed).
    """
    h, w = lop_shape
    lo_pth, lo_subpths = DistributedHDF5.create(
        os.path.join(dirpath, lo_fmt),
        num_outer_measurements,
        (w,),
        torch_dtype_as_str(lop_dtype),
        # this needs to be transposed later, but we keep it tall to allow
        # for QR without having to transpose (which may load to RAM)
        filedim_last=True,
    )
    #
    if with_ro:
        ro_pth, ro_subpths = DistributedHDF5.create(
            os.path.join(dirpath, ro_fmt),
            num_outer_measurements,
            (h,),
            torch_dtype_as_str(lop_dtype),
            filedim_last=True,
        )
    else:
        ro_pth, ro_subpths = None, None
    #
    c_pth, c_subpths = DistributedHDF5.create(
        os.path.join(dirpath, inner_fmt),
        num_inner_measurements,
        (num_inner_measurements,),
        torch_dtype_as_str(lop_dtype),
        filedim_last=True,
    )
    return ((lo_pth, lo_subpths), (ro_pth, ro_subpths), (c_pth, c_subpths))


# ##############################################################################
# # QR DECOMPOSITIONS
# ##############################################################################
def orthogonalize(matrix, overwrite=False):
    """Orthogonalization of given matrix.

    :param matrix: matrix to orthogonalize, needs to be compatible with either
      ``scipy.linalg.qr`` or ``torch.linalg.qr``.
    :param overwrite: If true, ``matrix[:] = Q`` will also be performed.
    :returns: Orthogonal matrix ``Q`` such that ``matrix = Q @ R`` as per the
      QR decomposition.
    """
    h, w = matrix.shape
    assert h >= w, "Only non-fat matrices supported!"
    #
    if isinstance(matrix, torch.Tensor):
        Q = torch.linalg.qr(matrix, mode="reduced")[0]
    else:
        Q = scipy.linalg.qr(
            matrix,
            mode="economic",
            pivoting=True,
        )[0]
    #
    if overwrite:
        matrix[:] = Q
    return Q


# ##############################################################################
# # CORE SOLVERS
# ##############################################################################
def core_pinv_solve(Q, randlop, target):
    """Pseudo-inverse step as part of solving the core sketch matrix.

    The core matrix can be expressed in terms of ``pinv(randlop @ Q) @ target``
    operations (see `[TYUC2019, 2.7] <https://arxiv.org/abs/1902.08651>`_).
    Given the ``target`` operator, this expression can be computed
    directly via least squares without having to explicitly compute the ``pinv``
    pseudo-inverse, which is typically faster and more stable.

    This function is used as a sub-routine in :func:`solve_core_ssvd` and
    :func:`solve_core_seigh`.

    :param Q: Matrix of shape ``(h, outer_measurements)`` containing either the
      left-outer or right-outer measurements.
    :param randlop: Linear operator used to perform the inner measurements.
      If ``Q`` contains right-outer measurements, this must be the left-inner
      operator, and vice versa. Expected shape is ``(num_inner, h)``.
    :returns: matrix equivalent to ``pinv(randlop @ Q) @ target``
    """
    h, outer = Q.shape
    inner, h2 = randlop.shape
    assert h == h2, "Inconsistent Q and randlop shapes!"
    assert target.shape[0] == inner, "Inconsistent randlop and target shapes!"
    #
    with_torch = isinstance(Q, torch.Tensor)
    dtype = torch_dtype_as_str(Q.dtype) if with_torch else str(Q.dtype)
    torch_dtype = getattr(torch, dtype)
    #
    if with_torch:
        buff = torch.zeros(inner, dtype=torch_dtype).to(Q.device)
        reduced = torch.empty((inner, outer), dtype=torch_dtype).to(Q.device)
        for i in range(inner):
            buff[i - 1] = 0
            buff[i] = 1
            reduced[i, :] = (buff @ randlop) @ Q
        result = torch.linalg.lstsq(reduced, target).solution
    else:
        buff = torch.zeros(inner, dtype=torch_dtype)
        reduced = np.empty((inner, outer), dtype=dtype)
        for i in range(inner):
            buff[i - 1] = 0
            buff[i] = 1
            reduced[i, :] = (buff @ randlop).numpy() @ Q
        result = scipy.linalg.lstsq(reduced[:], target[:])[0]
    #
    return result


def solve_core_ssvd(ro_Q, lo_Q, inner_measurements, li_randlop, ri_randlop):
    """Solving the core sketch matrix for sketched SVD.

    Given the orthogonalized right- and left-outer measurements, as well as
    the inner measurements and the linear operators used to perform said
    random measurements in the context of a sketched SVD, this function solves
    the core matrix, following
    `[TYUC2019, 2.7] <https://arxiv.org/abs/1902.08651>`_.

    .. note::
      This implementation is compatible with torch tensors, numpy arrays and
      HDF5 arrays.

    :param ro_Q: Right-outer measurement matrix. Expected to be in thin
      vertical shape, i.e. ``(lop_height, num_outer_measurements)``. Also
      expected to be orthogonal, i.e. ``ro_Q.T @ ro_Q = I``.
    :param lo_Q: Left-outer measurement matrix. Expected to be in thin
      vertical shape, i.e. ``(lop_width, num_outer_measurements)``. Also
      expected to be orthogonal, i.e. ``lo_Q.T @ lo_Q = I``.
    :param inner_measurements: Inner measurement matrix of shape
      ``(num_inner, num_inner)``.
    :parma li_randlop: Torch linear operator used to perform inner measurements
      via ``li_randlop @ lop @ ri_randlop.T``. Expected shape is
      ``(inner_measurements, lop_height)``.
    :parma ri_randlop: see ``li_randlop``. Expected shape is
      ``(inner_measurements, lop_height)``.
    """
    # figure out if we are using tensors
    ro_tensor = isinstance(ro_Q, torch.Tensor)
    lo_tensor = isinstance(lo_Q, torch.Tensor)
    inner_tensor = isinstance(inner_measurements, torch.Tensor)
    assert (
        ro_tensor == lo_tensor == inner_tensor
    ), "Either all or none of the measurements should be tensors!"
    with_torch = ro_tensor
    # consistency of dtypes
    assert ro_Q.dtype == lo_Q.dtype, "Inconsistent ro_Q and lo_Q dtypes!"
    assert (
        ro_Q.dtype == inner_measurements.dtype
    ), "Inconsistent ro_Q and inner_measurement dtypes!"
    # consistency of outer measurement shapes
    h, num_outer = ro_Q.shape
    w, _ = lo_Q.shape
    assert (
        num_outer == lo_Q.shape[1]
    ), "Mismatching left and right number of outer measurements!"
    # consistency of inner measurement shapes
    num_inner = li_randlop.shape[0]
    assert (
        ri_randlop.shape[0] == num_inner
    ), "Mismatching left and right number of inner measurements!"
    assert li_randlop.shape[1] == h, "Mismatching li_randlop and lop height!"
    assert ri_randlop.shape[1] == w, "Mismatching li_randlop and lop width!"
    # solve core via pseudoinverses
    core = core_pinv_solve(ro_Q, li_randlop, inner_measurements)
    core = core_pinv_solve(lo_Q, ri_randlop, core.T).T

    # compute and return SVD decomposition of core
    if with_torch:
        core_U, core_S, core_Vt = torch.linalg.svd(core)
    else:
        core_U, core_S, core_Vt = scipy.linalg.svd(
            core, overwrite_a=True, lapack_driver="gesvd"
        )
    return core_U, core_S, core_Vt


def solve_core_seigh(Q, inner, li_randlop, ri_randlop):
    """Solving the core sketch matrix for sketched Hermitian eigendecomposition.

    Given the orthogonalized outer measurements, as well as the inner
    measurements and the linear operators used to perform said
    inner measurements in the context of a sketched EIGH, this function solves
    the core matrix, following [...].

    .. note::
      This implementation is compatible with torch tensors, numpy arrays and
      HDF5 arrays.

    :param Q: outer measurement matrix (side doesn't matter since matrix is
      assumed to be Hermitian). Expected to be in thin vertical shape, i.e.
      ``(lop_dim, num_outer_measurements)``. Also expected to be orthogonal,
      i.e. ``Q.T @ Q = I``.
    :param inner: Inner measurement matrix of shape ``(num_inner, num_inner)``.
    :parma li_randlop: Torch linear operator used to perform inner measurements
      via ``li_randlop @ lop @ ri_randlop.T``. Expected shape is
      ``(inner_measurements, lop_height)``.
    :parma ri_randlop: see ``li_randlop``. Expected shape is
      ``(inner_measurements, lop_height)``.
    """
    # figure out if we are using tensors
    o_tensor = isinstance(Q, torch.Tensor)
    i_tensor = isinstance(inner, torch.Tensor)
    assert (
        o_tensor == i_tensor
    ), "Either all or none of the measurements should be tensors!"
    with_torch = o_tensor
    # consistency of dtypes
    assert Q.dtype == inner.dtype, "Inconsistent datatypes!"
    # consistency of outer measurement shapes
    h, num_outer = Q.shape
    inner_h, inner_w = inner.shape
    assert inner_h == inner_w, "Inner operator not square?"
    assert li_randlop.shape == (inner_h, h), "Mismatching shapes!"
    assert ri_randlop.shape == (inner_h, h), "Mismatching shapes!"
    # solve asymmetric core via SVD and 2 pseudoinverses
    if with_torch:
        inner_U, inner_S, inner_V = torch.linalg.svd(inner)
    else:
        inner_U, inner_S, inner_V = scipy.linalg.svd(inner)
    l_core = core_pinv_solve(Q, li_randlop, inner_U)
    del inner_U
    r_core = core_pinv_solve(Q, ri_randlop, inner_V.T)
    del inner_V
    core = l_core @ (DiagonalLinOp(inner_S) @ r_core.T)
    del inner_S
    # compute Hermitian eigendecomposition of core and sort recovered eigdec
    # in descending magnitude, since may be non-PSD
    if with_torch:
        core_S, core_U = torch.linalg.eigh(core)  # mat = U @ diag(S) @ U.T
        _, perm = abs(core_S).sort(descending=True)
        core_S, core_U = core_S[perm], core_U[:, perm]
    else:
        core_S, core_U = scipy.linalg.eigh(core)  # mat = U @ diag(S) @ U.T
        perm = np.argsort(abs(core_S))[::-1]
        core_S, core_U = core_S[perm], core_U[:, perm]
    return core_U, core_S



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for a-posteriori functionality."""

import pytest
import torch

from skerch.a_posteriori import a_posteriori_error, scree_bounds
from skerch.synthmat import SynthMat

from . import exp_decay, rng_seeds, torch_devices  # noqa: F401


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def f64_rtol():
    """Relative error tolerance for float64."""
    result = {torch.float64: 1e-10}
    return result


@pytest.fixture
def f32_rtol():
    """Relative error tolerance for float32."""
    result = {torch.float32: 1e-3}
    return result


@pytest.fixture
def h_w_sym_measurements(request):
    """Test cases: (op_shape, is_symmetric, number_a_post_measurements)."""
    result = [
        ((10, 10), True, 50),
        ((10, 10), False, 50),
        ((10, 100), False, 50),
        ((100, 100), True, 50),
        ((100, 100), False, 50),
        ((100, 1000), False, 50),
        ((1000, 1000), True, 50),
        ((1000, 1000), False, 50),
    ]
    #
    if request.config.getoption("--quick"):
        result = result[:3]
    return result


@pytest.fixture
def mixture_weights(request):
    """Linear weights to perturb one random matrix with another."""
    result = [0.001, 0.1, 1]
    if request.config.getoption("--quick"):
        result = result[:2]
    return result


# ##############################################################################
# # A POSTERIORI ERROR ESTIMATION
# ##############################################################################
def test_posteriori_and_scree_correctness(
    rng_seeds,
    torch_devices,
    f64_rtol,
    f32_rtol,
    h_w_sym_measurements,
    exp_decay,
    mixture_weights,
):
    """Test case for a-posteriori error estimation and scree bounds.

    This test samples all kinds of exp-decaying matrices, and random
    perturbations of varying magnitude. Then compares the actual Frobenius
    of the residual with the estimated one. It also reuses part of the
    computations to extract and test the scree bounds. Tests:
    * The a-posteriori estimates are within 0.5x and 2x of the real value
    * The scree upper bounds are above the scree lower bounds
    * Scree bounds are non-ascending as rank progresses
    * Scree bounds get tighter as rank progresses
    """
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype, _rtol in {**f64_rtol, **f32_rtol}.items():
                for (h, w), sym, q in h_w_sym_measurements:
                    for dec in exp_decay:
                        for weight in mixture_weights:
                            for psd in (True, False):
                                # sample an exp matrix and its perturbation
                                mat1 = SynthMat.exp_decay(
                                    shape=(h, w),
                                    rank=1,
                                    decay=dec,
                                    symmetric=sym,
                                    seed=seed,
                                    dtype=dtype,
                                    device=device,
                                    psd=psd,
                                )
                                mat2 = mat1 + weight * SynthMat.exp_decay(
                                    shape=(h, w),
                                    rank=1,
                                    decay=dec,
                                    symmetric=sym,
                                    seed=seed + 1,
                                    dtype=dtype,
                                    device=device,
                                    psd=psd,
                                )
                                # measure their actual frobenius difference
                                residual = ((mat1 - mat2) ** 2).sum()
                                # a posteriori estimate of their difference
                                for adjoint in [True, False]:
                                    f1, f2, res_estimate = a_posteriori_error(
                                        mat1,
                                        mat2,
                                        q,
                                        seed=seed + 2,
                                        dtype=dtype,
                                        device=device,
                                        adjoint=adjoint,
                                    )[0]
                                    assert residual <= (
                                        res_estimate * 2
                                    ), "Too small a-posteriori estimate!"
                                    assert residual >= (
                                        res_estimate / 2
                                    ), "Too large a-posteriori estimate!"
                                # SCREE TEST:
                                if sym:
                                    S1 = torch.linalg.eigvalsh(mat1)
                                    S2 = torch.linalg.eigvalsh(mat2)
                                else:
                                    _, S1, _ = torch.linalg.svd(mat1)
                                    _, S2, _ = torch.linalg.svd(mat2)
                                _, perm = abs(S1).sort(descending=True)
                                S1 = S1[perm]
                                _, perm = abs(S2).sort(descending=True)
                                S2 = S2[perm]
                                scree_lo, scree_hi = scree_bounds(
                                    S2, f1**0.5, res_estimate**0.5
                                )
                                # check both bounds non-ascending, and hi>=lo
                                assert (
                                    scree_hi >= scree_lo
                                ).all(), "Scree upper bounds <= lower bounds!"
                                assert (
                                    scree_hi.diff() <= 0
                                ).all(), "Scree upper bounds are ascending!"
                                assert (
                                    scree_lo.diff() <= 0
                                ).all(), "Scree lower bounds are ascending!"
                                # check: as rank increases, bounds tighten
                                assert (
                                    (scree_hi - scree_lo).diff() <= 0
                                ).all(), "Scree bounds diverging with rank!"
                                # # Check that bounds surround the actual scree
                                # # curve. THIS IS NOT THE CASE FOR THIS TEST
                                # # CASE, BUT WE AREN'T DOING SVD SO IT MAY BE
                                # # WRONG. IGNORE
                                # import matplotlib.pyplot as plt

                                # actual_scree, _ = scree_bounds(
                                #     S1, (S1**2).sum() ** 0.5, 0
                                # )
                                # # plt.clf()
                                # # plt.plot(scree_lo)
                                # # plt.plot(scree_hi)
                                # # plt.plot(actual_scree, linestyle="dashed")
                                # # plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Utilities for distributed sketch measurements.

One typically expensive step when we want to do sketched decompositions is
taking measurements from the linear operator we want to decompose, either
in terms of runtime, memory, or both. This module provides functionality to
parallelize and distribute said measurements.
"""


import os

import h5py
import torch

from .ssrft import SSRFT


# ##############################################################################
# # DECOMPOSITION HELPERS
# ##############################################################################
class DistributedHDF5:
    """Class to manage HDF5 files holding distributed sketch measurements.

    In general, multiple processes are not allowed to concurrently open and
    write to HDF5 files. In order to allow for distributed, but coherent
    writing, this class allows to create many individual HDF5 files, and then
    'virtually' merge them into a single coherent dataset.

    Unfortunately, most OS don't allow a single process to open many files at
    once, and each sub-file here counts as one. As a result, loading such
    distributed HDF5 datasets will reach a breaking point, after which contents
    are not really being opened, and reported as "empty" (e.g. full of zeros).
    For those cases, this class also provides a ``merge_all`` method,
    to efficiently merge the distributed dataset into a monolithic one.

    In the virtual mode, all files are created in the same directory following
    a consistent naming convention, and this is expected to remain unchanged.
    Since this class is naturally intended for multiple concurrent
    processes/devices, it is designed in the form of a static class.

    Usage example (see :mod:`.distributed_decompositions` for more examples)::

      # create the empty separate HDF5 files on disk, and the "virtual" merger
      out_path = "/tmp/my_dataset_{}.h5"
      each_shape, num_files = (1000,), 5
      h5_path, h5_subpaths = DistributedHDF5.create(
        out_path, num_files, each_shape, torch_dtype_as_str(op_dtype)
      )

      # in a (possibly distributed & parallelized) loop, load and write parts
      for i in range(num_files):
          vals, flag, h5 = DistributedHDF5.load(h5_subpaths[i])
          vals[:] += 1
          flag[0] = 'done'
          h5.close()  # remember to close file handles when done!

      # merged data can be used as a (1000, 50) matrix by a separate machine
      all_data, all_flags, all_h5 = DistributedHDF5.load_virtual(h5_path)
      print(scipy.diag(all_data))
      print(all_flags[::2])
      all_h5.close()

      # convert virtual database into single monolithic one
      DistributedHDF5.merge_all(
          h5_path,
          delete_subfiles_while_merging=True,
      )
      all_data, all_flags, all_h5 = DistributedHDF5.load_virtual(h5_path)
      ...
    """

    MAIN_PATH = "ALL"
    SUBPATHS_FORMAT = "{0:010d}"
    DATA_NAME = "data"
    FLAG_NAME = "flags"
    FLAG_DTYPE = h5py.string_dtype()
    INITIAL_FLAG = "initialized"

    @classmethod
    def create(
        cls,
        base_path,
        num_files,
        shape,
        dtype,
        compression="lzf",
        filedim_last=True,
    ):
        """Create a distributed HDF5 measurement dataset.

        :param str base_path: Format string with the path to store created HFG5
          files, in the form ``<DIR>/my_dataset_{}.h5``.
        :param num_files: Number of HDF5 files to be created, each
          corresponding to one measurement.
        :param shape: Shape of the array inside each individual file. For
          linear measurements, this is a vector, e.g. ``(1000,)``.
        :param dtype: Datatype of the HDF5 arrays to be created.
        :param filedim_last: If true, the virtual dataset result of merging
          all files will be of shape ``shape + (num_files,)``. Otherwise,
          ``(num_files,) + shape``.
        :returns: The pair ``(all_path, subpaths)``, where ``all_path`` is the
          path of the virtual dataset encompassing all subpaths, which
          correspond to the individual files.
        """
        all_path = base_path.format(cls.MAIN_PATH)
        subpaths = [
            base_path.format(cls.SUBPATHS_FORMAT.format(i))
            for i in range(num_files)
        ]
        # create virtual dataset to hold everything together via softlinks
        # use relative paths to just assume everything is in the same dir
        if filedim_last:
            data_shape = shape + (num_files,)
        else:
            data_shape = (num_files,) + shape
        data_lyt = h5py.VirtualLayout(shape=data_shape, dtype=dtype)
        flag_lyt = h5py.VirtualLayout(shape=(num_files,), dtype=cls.FLAG_DTYPE)
        for i, p in enumerate(subpaths):
            p = os.path.basename(p)  # relative path
            vs = h5py.VirtualSource(p, cls.DATA_NAME, shape=shape)
            if filedim_last:
                data_lyt[..., i] = vs
            else:
                data_lyt[i] = vs
            flag_lyt[i] = h5py.VirtualSource(p, cls.FLAG_NAME, shape=(1,))
        all_h5 = h5py.File(all_path, "w")  # , libver="latest")
        all_h5.create_virtual_dataset(cls.DATA_NAME, data_lyt)
        all_h5.create_virtual_dataset(cls.FLAG_NAME, flag_lyt)
        all_h5.close()
        # create separate HDF5 files
        for p in subpaths:
            h5f = h5py.File(p, "w")
            h5f.create_dataset(
                cls.DATA_NAME,
                shape=shape,
                maxshape=shape,
                dtype=dtype,
                compression=compression,
                chunks=shape,
            )
            h5f.create_dataset(
                cls.FLAG_NAME,
                shape=(1,),
                maxshape=(1,),
                compression=compression,
                dtype=cls.FLAG_DTYPE,
                chunks=(1,),
            )
            h5f[cls.FLAG_NAME][:] = cls.INITIAL_FLAG
            h5f.close()
        #
        return all_path, subpaths

    @classmethod
    def load(cls, path, filemode="r+"):
        """Load an individual dataset created via :meth:`.create`.

        :param path: One of the subpaths returned by :meth:`.create`.
        :param filemode: Default is 'r+', read/write, file must preexist. See
          documentation of ``h5py.File`` for more details.
        :returns: ``(data, flag, h5f)``, where ``data`` is the dataset
          for the numerical measurements, ``flag`` is the dataset for state
          tracking, and ``h5f`` is the (open) HDF5 file handle.

        .. note::

          Remember to ``h5f.close()`` once done with this file.
        """
        h5f = h5py.File(path, filemode)
        data = h5f[cls.DATA_NAME]
        flag = h5f[cls.FLAG_NAME]
        return data, flag, h5f

    @classmethod
    def load_virtual(cls, all_path, filemode="r+"):
        """Load a merged dataset created via :meth:`.create`.

        :param path: The ``all_path`` returned by :meth:`.create`.
        :param filemode: Default is 'r+', read/write, file must preexist. See
          documentation of ``h5py.File`` for more details.
        :returns: ``(data, flag, h5f)``, where ``data`` is the dataset
          for the numerical measurements, ``flag`` is the dataset for state
          tracking, and ``h5f`` is the (open) HDF5 file handle.

        .. note::

          Remember to ``h5f.close()`` once done with this file.
        """
        all_h5f = h5py.File(all_path, filemode)
        data = all_h5f[cls.DATA_NAME]
        flags = all_h5f[cls.FLAG_NAME]
        return data, flags, all_h5f

    @classmethod
    def analyze_virtual(cls, all_path, check_success_flag=None):
        """Analyze shapes and relations of files created via :meth:`.create`.

        Extract relevant information from a given virtual HDF5 dataset, which
        can be used e.g. to allow merging it into a monolithic one (see
        implementation of :meth:`merge_all` for an example).

        :param all_path: The ``all_path`` of a virtual HDF5 dataset like the
          ones created via :meth:`.create`.
        :param check_success_flag: If given, this method will check that all
          HDF5 flags equal this value, raise an ``AssertionError`` otherwise.
        :returns: the tuple ``(data_shape, data_dtype, data_subshapes,
          data_map, flags_shape, flags_dtype, flag_subshapes, flag_map,
          filedim_idx)``.
        """
        # load virtual HDF5 and grab main infos. Note that we avoid
        # using data and flags for anything else, since they may be affected by
        # the "too many open files" OS issue.
        data, flags, h5f = DistributedHDF5.load_virtual(all_path, filemode="r")
        data_shape, flags_shape = data.shape, flags.shape
        data_dtype, flags_dtype = data.dtype, flags.dtype
        abspath = h5f.filename
        rootdir = os.path.dirname(abspath)
        # figure out involved paths and their respective indices in virtual
        data_map, flag_map = {}, {}
        data_subshapes, flag_subshapes = [], []
        for vs in h5f[cls.DATA_NAME].virtual_sources():
            subpath = os.path.join(rootdir, vs.file_name)
            begs_ends = tuple(
                (b, e + 1) for b, e in zip(*vs.vspace.get_select_bounds())
            )
            shape = tuple(e - b for b, e in begs_ends)
            data_map[begs_ends] = subpath
            data_subshapes.append(shape)
        for vs in h5f[cls.FLAG_NAME].virtual_sources():
            subpath = os.path.join(rootdir, vs.file_name)
            begs_ends = tuple(
                (b, e + 1) for b, e in zip(*vs.vspace.get_select_bounds())
            )
            shape = tuple(e - b for b, e in begs_ends)
            assert shape == (1,), "Flags expected to have shape (1,)!"
            flag_map[begs_ends] = subpath
            flag_subshapes.append(shape)
        subpaths = set(data_map.values())
        # figure out position of filedim was first or last
        is_filedim = [
            (a - b) == 0 for a, b in zip(data_shape, data_subshapes[0])
        ]
        assert sum(is_filedim) == 1, "Only one running dimension supported!"
        filedim_idx = is_filedim.index(False)
        # virtual sanity check and close virtual
        data_beginnings = {k[filedim_idx][0] for k in data_map.keys()}
        assert len(data_beginnings) == len(
            h5f[cls.DATA_NAME].virtual_sources()
        ), "Repeated file_idx beginnings in data?"
        assert len(flag_map) == len(
            h5f[cls.FLAG_NAME].virtual_sources()
        ), "Repeated indices in flags?"
        for sp in subpaths:
            assert os.path.isfile(sp), f"Subpath doesn't exist! {sp}"
        for sp2 in flag_map.values():
            assert sp2 in subpaths, "Flag subpaths different to data subpath!"
        assert (
            len(set(data_subshapes)) == 1
        ), "Heterogeneous data shapes in virtual dataset not supported!"
        assert (
            len(set(flag_subshapes)) == 1
        ), "Heterogeneous flag shapes in virtual dataset not supported!"
        h5f.close()
        # if success flag was given, check every individual sub-HDF5 was
        # successful
        if check_success_flag is not None:
            for sp in subpaths:
                _, flag, h5 = cls.load(sp, filemode="r")
                flag = flag[0].decode()
                assert (
                    flag == check_success_flag
                ), f"Unsuccessful flag in {sp}! {flag}"
                h5.close()
        #
        return (
            data_shape,
            data_dtype,
            data_subshapes,
            data_map,
            flags_shape,
            flags_dtype,
            flag_subshapes,
            flag_map,
            filedim_idx,
        )

    @classmethod
    def merge_all(
        cls,
        all_path,
        out_path=None,
        compression="lzf",
        check_success_flag=None,
        delete_subfiles_while_merging=False,
    ):
        """Merges distributed HDF5 dataset into a single, monolithic HDF5 file.

        :param all_path: The ``all_path`` of a virtual HDF5 dataset like the
          ones created via :meth:`.create`.
        :param out_path: If None, merged dataset will be written over the given
          ``all_path``. Otherwise, path of the resulting HDF5 monolithic file.
        :param check_success_flag: If given, this method will check that all
          HDF5 flags equal this value, raise an ``AssertionError`` otherwise.
        :param bool delete_subfiles_while_merging: If true, each distributed
          HDF5 file that is visited will be deleted right after it is merged
          onto the monolithic HDF5 file. Useful to avoid large memory
          overhead.
        :returns: ``out_path``.
        """
        # grab relevant infos from virtual, we still didn't modify anything
        (
            data_shape,
            data_dtype,
            data_subshapes,
            data_map,
            flags_shape,
            flags_dtype,
            flag_subshapes,
            flag_map,
            filedim_idx,
        ) = cls.analyze_virtual(all_path, check_success_flag)
        # create HDF5 that we will iteratively expand. If no out_path,
        # we will overwrite virtual.
        if out_path is None:
            out_path = all_path
        h5f = h5py.File(out_path, "w")
        #
        init_shape = list(data_shape)
        init_shape[filedim_idx] = 0
        h5f.create_dataset(
            cls.DATA_NAME,
            shape=init_shape,
            maxshape=data_shape,
            dtype=data_dtype,
            compression=compression,
            chunks=data_subshapes[0],
        )
        h5f.create_dataset(
            cls.FLAG_NAME,
            shape=0,
            maxshape=flags_shape,
            dtype=flags_dtype,
            compression=compression,
            chunks=flag_subshapes[0],
        )
        # iterate over contents in sorted order and extend h5f with them
        sorted_data = sorted(data_map, key=lambda x: x[filedim_idx][0])
        for begs_ends in sorted_data:
            subpath = data_map[begs_ends]
            subdata, subflag, h5 = cls.load(subpath, filemode="r")
            if check_success_flag is not None:
                assert (
                    subflag[0].decode() == check_success_flag
                ), f"Subfile flag not equal {check_success_flag}!"
            # increment size of h5f by 1 entry
            data_shape = list(h5f[cls.DATA_NAME].shape)
            data_shape[filedim_idx] += 1
            h5f[cls.DATA_NAME].resize(data_shape)
            h5f[cls.FLAG_NAME].resize((len(h5f[cls.FLAG_NAME]) + 1,))
            # write subdata and subflags to h5f, flush and close subfile
            target_slices = tuple(slice(*be) for be in begs_ends)
            h5f[cls.DATA_NAME][target_slices] = subdata[:].reshape(
                data_subshapes[0]
            )
            h5f[cls.FLAG_NAME][-1:] = subflag[:]
            h5f.flush()
            h5.close()
            # optionally, delete subfile
            if delete_subfiles_while_merging:
                os.remove(subpath)
        #
        h5f.close()
        return out_path


# ##############################################################################
# # DISTRIBUTED MEASUREMENTS
# ##############################################################################
def linmeas_idx_torch(idx, lop, meas_lop, lop_device, lop_dtype, adjoint=False):
    """Distributed-able linear measurement from given linear operators.

    Convenience function to perform a single linear measurement from two
    (potentially matrix-free) linear operators. Specifically, it picks the
    ``[idx]`` row of ``meas_lop``, and applies it to ``lop``.

    :param lop: Linear operator to measure from, typically matrix-free or very
      costly to store or compute. Assumed to implement right- and left matrix
      multiplication via the ``@`` operator, and also the ``shape = (h, w)``
      attribute, but no transposition.
    :param meas_lop: Linear operator to be applied to ``lop``. It must also
      support ``@`` operator and ``shape`` attribute, but not transposition is
      required. For this reason, its shape is ``(*, h)`` if adjoint is true,
      and ``(*, w)`` otherwise, where ``*`` is arbitrary.
    :returns: If adjoint is True, ``meas_lop[idx, :] @ lop``. Otherwise
      ``lop @ meas_lop[idx, :].T``.
    """
    rh, rw = meas_lop.shape
    v = torch.zeros(rh, dtype=lop_dtype).to(lop_device)
    v[idx] = 1
    try:
        if adjoint:
            result = (v @ meas_lop) @ lop
        else:
            result = lop @ (v @ meas_lop)
    except TypeError as te:
        msg = "Error during matmul! LinOp expected to work with torch tensors."
        raise TypeError(msg) from te
    return result


def ssrft_idx_torch(
    idx,
    total_measurements,
    lop,
    lop_device,
    lop_dtype,
    out_vals,
    seed=0b1110101001010101011,
    flag=None,
    processing_flag="-1",
    success_flag="1",
    adjoint=False,
):
    """Distributed-able SSRFT measurement from given linear operators.

    .. note::

      See :mod:`.ssrft` for more details on the SSRFT and its implementation.

    In-place SSRFT measurement, it projects the ``idx`` row of a SSRFT operator
    onto the given ``lop``, writing results to ``out_vals``. If adjoint is
    true, computes ``ssrft[idx, :] @ lop``, otherwise ``lop @ ssrft[idx, :].T``,
    where ``ssrft``is always of shape ``(total_measurements, *)`` with ``*``
    being the fitting dimension of ``lop``.

    :param out_vals: Array where the results will be writen in-place, of shape
      ``(*, )`` for either the height or width of ``lop``. It can be a ``torch``
      tensor or a ``numpy`` or a ``HDF5`` array, accepting in-place  writing
      via ``out_vals[:] = ...``.
    :param flag: An array-like object of shape (1,) and string datatype,
      optionally used by this function to keep track of current state by
      setting it to the ``processing_flag, success_flag`` values via
      the ``flag[0] = "..."`` assignment.
    :param bool adjoint: See above.
    :returns: the SSRFT matrix-free linear operator used for the measurement.
      Note that the measurement is not returned, but written to ``out_vals``,
      and optionally ``flag`` is also written.
    """
    assert (
        0 <= idx <= total_measurements
    ), "OOB index for number of measurements!"
    vals_shape = out_vals.shape
    lop_h, lop_w = lop.shape
    if adjoint:
        ssrft = SSRFT((total_measurements, lop_h), seed=seed)
        assert vals_shape == (lop_w,), "Mismatching measurement and out_shapes!"
    else:
        ssrft = SSRFT((total_measurements, lop_w), seed=seed)
        assert vals_shape == (lop_h,), "Mismatching measurement and out_shapes!"
    #
    if flag is not None:
        flag[0] = processing_flag
    #
    result = linmeas_idx_torch(idx, lop, ssrft, lop_device, lop_dtype, adjoint)
    try:
        # try to assign as-is (covers torch-to-torch and CPU cases)
        out_vals[:] = result
    except Exception:
        # if anything happens, try to copy to CPU
        out_vals[:] = result.cpu()
    if flag is not None:
        flag[0] = success_flag
    return ssrft


# ##############################################################################
# # DISTRIBUTED INNER MEASUREMENTS
# ##############################################################################
def innermeas_idx_torch(
    idx,
    num_measurements,
    lop,
    lop_device,
    lop_dtype,
    out,
    left_seed,
    right_seed,
    flag=None,
    processing_flag="0",
    success_flag="1",
):
    """Distributed-able SSRFT inner measurement for sketched decompositions.

    .. note::

      See :mod:`.ssrft` for more details on the SSRFT and its implementation.

    Like :func:`ssrft_idx_torch`, but it applies the SSRFT on both sides of
    ``lop``, i.e. it computes: ``ssrft_l @ lop @ ssrft_r[idx, :].T``, where
    ``ssrft_l`` has ``left_seed`` and shape ``(num_measurements, lop.h)``, and
    ``ssrft_r`` has ``right_seed`` and shape ``(num_measurements, lop.w)``.

    :param out: Array where the results will be writen in-place, of shape
      ``(num_measurements, )``. It can be a ``torch`` tensor or a ``numpy`` or
      a ``HDF5`` array, accepting in-place  writing via ``out[:] = ...``.
    :param flag: See :func:`ssrft_idx_torch`.
    :returns: The in-between result ``lop @ ssrft_r[idx, :].T``, as a torch
      tensor of shape ``(lop_height,)``. It will be on CUDA device if
      ``lop_device`` is CUDA, otherwise CPU.
    """
    h, w = lop.shape
    #
    li_ssrft = SSRFT((num_measurements, h), seed=left_seed)
    buff = torch.zeros(h, dtype=lop_dtype).to(lop_device)
    ssrft_idx_torch(
        idx,
        num_measurements,
        lop,
        lop_device,
        lop_dtype,
        buff[:],
        right_seed,
        flag,
        processing_flag=processing_flag,
        success_flag=success_flag,
        adjoint=False,
    )
    if flag is not None:
        flag[0] = processing_flag
    try:
        # try to assign as-is (covers torch-to-torch and CPU cases)
        out[:] = li_ssrft @ buff
    except Exception:
        # if anything happens, try to copy to CPU
        out[:] = (li_ssrft @ buff).cpu()
        buff = buff.cpu()
    if flag is not None:
        flag[0] = success_flag
    return buff



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for QR decomposition."""


import pytest
import torch

from skerch.utils import gaussian_noise

from . import rng_seeds, torch_devices


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def heights_widths():
    """Shapes of matrices to QR decompose."""
    result = [
        (1, 1),
        (10, 10),
        (100, 100),
        (1_000, 1_000),
    ]
    result += [
        (10, 1),
        (100, 10),
        (1_000, 100),
        (10_000, 1_000),
    ]
    return result


@pytest.fixture
def entry_atol():
    """Absolute, entry-wise error tolerance for ``diff(Q.T @ Q, I)``."""
    result = 1e-5
    return result


# ##############################################################################
# #
# ##############################################################################
def test_orth_q(rng_seeds, torch_devices, heights_widths, entry_atol):
    """Test case for orthogonality of Q in QR decomposition.

    Tests that ``Q.T @ Q`` is close to identity.
    """
    for h, w in heights_widths:
        assert h >= w, "This test doesn't need/admit fat matrices!"
    #
    for seed in rng_seeds:
        for device in torch_devices:
            for dtype in (torch.float64, torch.float32):
                for h, w in heights_widths:
                    mat = gaussian_noise(
                        (h, w),
                        mean=0.0,
                        std=1.0,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                    )
                    Q = torch.linalg.qr(mat)[0]
                    I_residual = Q.T @ Q
                    I_residual[range(w), range(w)] -= 1
                    worst = I_residual.abs().max().item()
                    assert worst <= abs(entry_atol), "Q matrix not orthogonal?"



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for distributed sketched matrix decompositions."""


import tempfile

import pytest
import torch

from skerch.a_posteriori import a_posteriori_error
from skerch.decompositions import truncate_core
from skerch.synthmat import SynthMat

from . import rng_seeds, torch_devices  # noqa: F401
from .dseigh import mock_dseigh
from .dssvd import mock_dssvd


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def atol_exp():
    """Absolute error tolerances for exp-decaying test matrices."""
    result = {torch.float64: 1e-7, torch.float32: 1e-3}
    return result


@pytest.fixture
def atol_poly():
    """Absolute error tolerances for poly-decaying test matrices."""
    result = {torch.float64: 3e-3, torch.float32: 3e-3}
    return result


@pytest.fixture
def heights_widths_ranks_outer_inner_posteriori(request):
    """Test cases for different matrices and measurements.

    Entries are ``((h, w), r, (o, i)), q``. For a matrix of shape ``h, w`` and
    rank ``r``, the Skinny SVD will do ``o`` outer measurements and ``i``
    inner measurements. Recommended is that ``i >= 2*o``. The a-posteriori
    number of measurements is ``q``.

    For constrained budgets, the Skinny SVD will naturally yield higher error
    with smaller shapes, hence we test a few medium shapes.
    """
    result = [
        ((1_000, 1_000), 10, (100, 300), 30),
        ((1_000, 1_000), 50, (200, 600), 30),
        # ((1_000, 2_000), 10, (100, 300), 30),  # commentd to save time
        ((1_000, 2_000), 50, (200, 600), 30),
        # ((2_000, 2_000), 20, (100, 300), 30),  # commented to save time
        ((2_000, 2_000), 100, (200, 600), 30),
    ]
    if request.config.getoption("--quick"):
        result = result[1:2]
    return result


@pytest.fixture
def dims_ranks_outer_inner_posteriori(request):
    """Test cases for different (square) matrices and measurements.

    Like ``heights_widths_ranks_outer_inner_posteriori`` but for square shapes
    to test on symmetric matrices.
    """
    result = [
        (1_000, 10, (100, 300), 30),
        (1_000, 25, (200, 600), 30),
        # (2_000, 10, (100, 300), 30),  # commented to save time
        (2_000, 50, (200, 600), 30),
        (2_000, 100, (200, 600), 30),
    ]
    if request.config.getoption("--quick"):
        result = result[1:2]
    return result


@pytest.fixture
def recons_frob_tol():
    """Ratio tolerances for Frobenius residual errors of recovered matrices.

    Threshold for the Frobenius of residual divided by energy of original.
    If 0.05, it means that the residual must be 5% or less of the original.
    """
    result = {torch.float64: 0.05, torch.float32: 0.05}
    return result


@pytest.fixture
def svectors_frob_tol():
    """Ratio tolerances for Frobenius residual errors of recovered eigenbases.

    Threshold for the Frobenius of residual divided by energy of original.
    If 0.05, it means that the residual must be 5% or less of the original.
    """
    result = {torch.float64: 0.05, torch.float32: 0.05}
    return result


@pytest.fixture
def svectors_check_n():
    """How many singular vectors to check."""
    return 5


@pytest.fixture
def high_snr(request):
    """SNR values for Lowrank+noise matrix. The larger, the more noise.

    These values are low-noise, allowing for better recoveries.
    """
    if request.config.getoption("--quick"):
        return [1e-2]
    else:
        return [1e-3, 1e-2]


@pytest.fixture
def steep_exp_decay(request):
    """Steep exponential decay values, which allow for better recoveries.

    The larger, the faster decay.
    """
    if request.config.getoption("--quick"):
        return [0.1]
    else:
        return [0.5, 0.1]


@pytest.fixture
def steep_poly_decay(request):
    """Steep polynomial decay values, which allow for better recoveries.

    The larger, the faster decay.
    """
    if request.config.getoption("--quick"):
        return [2]
    else:
        return [4, 2]


# ##############################################################################
# # SKETCHED SVD
# ##############################################################################
def test_dssvd_asymmetric_exp(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    atol_exp,
    heights_widths_ranks_outer_inner_posteriori,
    steep_exp_decay,
    recons_frob_tol,
    svectors_frob_tol,
    svectors_check_n,
):
    """Test distributed SSVD on asymmetric exponential matrices.

    Sample matrices, and compute their ``torch.svd`` as well as the sketched
    SVD. Test that:
    * Recovered singular values are correct
    * Recovered matrix is correct
    * The first bulk singular vectors are correct
    * A-posteriori error estimation is low
    """
    for seed in rng_seeds:
        for dtype, atol in atol_exp.items():
            for (
                (h, w),
                r,
                (outer, inner),
                post,
            ) in heights_widths_ranks_outer_inner_posteriori:
                for dec in steep_exp_decay:
                    # we create matrix and full SVD on cpu, it is faster
                    mat = SynthMat.exp_decay(
                        shape=(h, w),
                        rank=r,
                        decay=dec,
                        symmetric=False,
                        seed=seed,
                        dtype=dtype,
                        device="cpu",
                    )
                    U, S, Vt = torch.linalg.svd(mat)
                    # then we test decomposition on both devices
                    for device in torch_devices:
                        mat = mat.to(device)
                        U = U.to(device)
                        S = S.to(device)
                        Vt = Vt.to(device)
                        #
                        tmpdir = tempfile.TemporaryDirectory()
                        (ro_Q, core_U, core_S, core_Vt, lo_Qt) = mock_dssvd(
                            tmpdir.name, mat, device, dtype, outer, inner, seed
                        )
                        # check that singular values are correct
                        assert torch.allclose(
                            S[: 2 * r], core_S[: 2 * r], atol=atol
                        ), "Bad recovery of singular values!"
                        # check that recovered matrix is the same
                        core_U, core_S, core_Vt = truncate_core(
                            2 * r, core_U, core_S, core_Vt
                        )
                        mat_recons = (
                            ro_Q @ core_U @ torch.diag(core_S) @ core_Vt @ lo_Qt
                        )
                        recons_err = torch.dist(mat, mat_recons) / mat.norm()
                        assert recons_err <= abs(
                            recons_frob_tol[dtype]
                        ), "Bad SSVD reconstruction!"
                        # check that a few singular vectors are correct: avoid
                        # the ones that have the same singular values, since
                        # they aren't unique. Also compare outer prods, since
                        # multiplying left and right by -1 yields same result.
                        left = ro_Q @ core_U[:, r:]
                        right = core_Vt[r:, :] @ lo_Qt

                        for idx in range(svectors_check_n):
                            recons_err = torch.dist(
                                torch.outer(left[:, idx], right[idx]),
                                torch.outer(U[:, idx + r], Vt[idx + r]),
                            ) / (U[:, idx + r].norm() * Vt[idx + r].norm())
                            assert recons_err <= abs(
                                svectors_frob_tol[dtype]
                            ), "Bad recovery of singular vectors!"
                        # check a-posteriori error. With 30 measurements, we
                        # know that the probability of the actual error being
                        # twice as big is less than 1%, and keeps shrinking
                        # exponentially: a_posteriori_error_bounds(30, 1)
                        # Therefore, a very small estimate almost certainly
                        # guarantees that the approximation is good, and we
                        # have enough rank in our truncated SVD.
                        for adjoint in [True, False]:
                            _, _, err_estimate = a_posteriori_error(
                                mat,
                                mat_recons,
                                post,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                adjoint=adjoint,
                            )[0]
                            assert (
                                err_estimate / mat.numel()
                            ) <= atol, "Bad a-posteriori estimate!"
                        # cleanup HDF5 tmpdir
                        tmpdir.cleanup()


def test_dssvd_symmetric_exp(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    atol_exp,
    dims_ranks_outer_inner_posteriori,
    steep_exp_decay,
    recons_frob_tol,
    svectors_frob_tol,
    svectors_check_n,
):
    """Test distributed SSVD on symmetric exponential matrices.

    Sample matrices, and compute their ``torch.svd`` as well as the sketched
    SVD. Test that:
    * Recovered singular values are correct
    * Recovered matrix is correct and symmetric
    * The first bulk singular vectors are correct
    * A-posteriori error estimation is low
    """
    for seed in rng_seeds:
        for dtype, atol in atol_exp.items():
            for d, r, (outer, inner), post in dims_ranks_outer_inner_posteriori:
                for dec in steep_exp_decay:
                    # we create matrix and full SVD on cpu, it is faster
                    mat = SynthMat.exp_decay(
                        shape=(d, d),
                        rank=r,
                        decay=dec,
                        symmetric=True,
                        seed=seed,
                        dtype=dtype,
                        device="cpu",
                    )
                    U, S, Vt = torch.linalg.svd(mat)

                    # then we test decomposition on both devices
                    for device in torch_devices:
                        mat = mat.to(device)
                        U = U.to(device)
                        S = S.to(device)
                        Vt = Vt.to(device)
                        #
                        tmpdir = tempfile.TemporaryDirectory()
                        (ro_Q, core_U, core_S, core_Vt, lo_Qt) = mock_dssvd(
                            tmpdir.name, mat, device, dtype, outer, inner, seed
                        )
                        # check that singular values are correct
                        assert torch.allclose(
                            S[: 2 * r], core_S[: 2 * r], atol=atol
                        ), "Bad recovery of singular values!"
                        # check that recovered matrix is the same
                        core_U, core_S, core_Vt = truncate_core(
                            2 * r, core_U, core_S, core_Vt
                        )
                        mat_recons = (
                            ro_Q @ core_U @ torch.diag(core_S) @ core_Vt @ lo_Qt
                        )
                        recons_err = torch.dist(mat, mat_recons) / mat.norm()
                        assert recons_err <= abs(
                            recons_frob_tol[dtype]
                        ), "Bad SSVD reconstruction!"
                        # check that a few singular vectors are correct: avoid
                        # the ones that have the same singular values, since
                        # they aren't unique. Also compare outer prods, since
                        # multiplying left and right by -1 yields same result.
                        left = ro_Q @ core_U[:, r:]
                        right = core_Vt[r:, :] @ lo_Qt

                        for idx in range(svectors_check_n):
                            recons_err = torch.dist(
                                torch.outer(left[:, idx], right[idx]),
                                torch.outer(U[:, idx + r], Vt[idx + r]),
                            ) / (U[:, idx + r].norm() * Vt[idx + r].norm())
                            assert recons_err <= abs(
                                svectors_frob_tol[dtype]
                            ), "Bad recovery of singular vectors!"
                        # check a-posteriori error. With 30 measurements, we
                        # know that the probability of the actual error being
                        # twice as big is less than 1%, and keeps shrinking
                        # exponentially: a_posteriori_error_bounds(30, 1)
                        # Therefore, a very small estimate almost certainly
                        # guarantees that the approximation is good, and we
                        # have enough rank in our truncated SVD.
                        for adjoint in [True, False]:
                            _, _, err_estimate = a_posteriori_error(
                                mat,
                                mat_recons,
                                post,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                adjoint=adjoint,
                            )[0]
                            assert (
                                err_estimate / mat.numel()
                            ) <= atol, "Bad a-posteriori estimate!"
                        # cleanup HDF5 tmpdir
                        tmpdir.cleanup()


def test_dssvd_asymmetric_poly(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    atol_poly,
    heights_widths_ranks_outer_inner_posteriori,
    steep_poly_decay,
    recons_frob_tol,
    svectors_frob_tol,
    svectors_check_n,
):
    """Test distributed SSVD on asymmetric polynomial matrices.

    Sample matrices, and compute their ``torch.svd`` as well as the sketched
    SVD. Test that:
    * Recovered singular values are correct
    * Recovered matrix is correct
    * The first bulk singular vectors are correct
    * A-posteriori error estimation is low
    """
    for seed in rng_seeds:
        for dtype, atol in atol_poly.items():
            for (
                (h, w),
                r,
                (outer, inner),
                post,
            ) in heights_widths_ranks_outer_inner_posteriori:
                for dec in steep_poly_decay:
                    # we create matrix and full SVD on cpu, it is faster
                    mat = SynthMat.poly_decay(
                        shape=(h, w),
                        rank=r,
                        decay=dec,
                        symmetric=False,
                        seed=seed,
                        dtype=dtype,
                        device="cpu",
                    )
                    U, S, Vt = torch.linalg.svd(mat)
                    # then we test decomposition on both devices
                    for device in torch_devices:
                        mat = mat.to(device)
                        U = U.to(device)
                        S = S.to(device)
                        Vt = Vt.to(device)
                        #
                        tmpdir = tempfile.TemporaryDirectory()
                        (ro_Q, core_U, core_S, core_Vt, lo_Qt) = mock_dssvd(
                            tmpdir.name, mat, device, dtype, outer, inner, seed
                        )
                        # check that singular values are correct
                        assert torch.allclose(
                            S[: 2 * r], core_S[: 2 * r], atol=atol
                        ), "Bad recovery of singular values!"
                        # check that recovered matrix is the same
                        core_U, core_S, core_Vt = truncate_core(
                            2 * r, core_U, core_S, core_Vt
                        )
                        mat_recons = (
                            ro_Q @ core_U @ torch.diag(core_S) @ core_Vt @ lo_Qt
                        )
                        recons_err = torch.dist(mat, mat_recons) / mat.norm()
                        assert recons_err <= abs(
                            recons_frob_tol[dtype]
                        ), "Bad SSVD reconstruction!"
                        # check that a few singular vectors are correct: avoid
                        # the ones that have the same singular values, since
                        # they aren't unique. Also compare outer prods, since
                        # multiplying left and right by -1 yields same result.
                        left = ro_Q @ core_U[:, r:]
                        right = core_Vt[r:, :] @ lo_Qt

                        for idx in range(svectors_check_n):
                            recons_err = torch.dist(
                                torch.outer(left[:, idx], right[idx]),
                                torch.outer(U[:, idx + r], Vt[idx + r]),
                            ) / (U[:, idx + r].norm() * Vt[idx + r].norm())
                            assert recons_err <= abs(
                                svectors_frob_tol[dtype]
                            ), "Bad recovery of singular vectors!"
                        # check a-posteriori error. With 30 measurements, we
                        # know that the probability of the actual error being
                        # twice as big is less than 1%, and keeps shrinking
                        # exponentially: a_posteriori_error_bounds(30, 1)
                        # Therefore, a very small estimate almost certainly
                        # guarantees that the approximation is good, and we
                        # have enough rank in our truncated SVD.
                        for adjoint in [True, False]:
                            _, _, err_estimate = a_posteriori_error(
                                mat,
                                mat_recons,
                                post,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                adjoint=adjoint,
                            )[0]
                            assert (
                                err_estimate / mat.numel()
                            ) <= atol, "Bad a-posteriori estimate!"
                        # cleanup HDF5 tmpdir
                        tmpdir.cleanup()


def test_dssvd_symmetric_poly(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    atol_poly,
    dims_ranks_outer_inner_posteriori,
    steep_poly_decay,
    recons_frob_tol,
    svectors_frob_tol,
    svectors_check_n,
):
    """Test distributed SSVD on symmetric polynomial matrices.

    Sample matrices, and compute their ``torch.svd`` as well as the sketched
    SVD. Test that:
    * Recovered singular values are correct
    * Recovered matrix is correct and symmetric
    * The first bulk singular vectors are correct
    * A-posteriori error estimation is low
    """
    for seed in rng_seeds:
        for dtype, atol in atol_poly.items():
            for d, r, (outer, inner), post in dims_ranks_outer_inner_posteriori:
                for dec in steep_poly_decay:
                    # we create matrix and full SVD on cpu, it is faster
                    mat = SynthMat.poly_decay(
                        shape=(d, d),
                        rank=r,
                        decay=dec,
                        symmetric=True,
                        seed=seed,
                        dtype=dtype,
                        device="cpu",
                    )
                    U, S, Vt = torch.linalg.svd(mat)

                    # then we test decomposition on both devices
                    for device in torch_devices:
                        mat = mat.to(device)
                        U = U.to(device)
                        S = S.to(device)
                        Vt = Vt.to(device)
                        #
                        tmpdir = tempfile.TemporaryDirectory()
                        (ro_Q, core_U, core_S, core_Vt, lo_Qt) = mock_dssvd(
                            tmpdir.name, mat, device, dtype, outer, inner, seed
                        )
                        # check that singular values are correct
                        assert torch.allclose(
                            S[: 2 * r], core_S[: 2 * r], atol=atol
                        ), "Bad recovery of singular values!"
                        # check that recovered matrix is the same
                        core_U, core_S, core_Vt = truncate_core(
                            2 * r, core_U, core_S, core_Vt
                        )
                        mat_recons = (
                            ro_Q @ core_U @ torch.diag(core_S) @ core_Vt @ lo_Qt
                        )
                        recons_err = torch.dist(mat, mat_recons) / mat.norm()
                        assert recons_err <= abs(
                            recons_frob_tol[dtype]
                        ), "Bad SSVD reconstruction!"
                        # check that a few singular vectors are correct: avoid
                        # the ones that have the same singular values, since
                        # they aren't unique. Also compare outer prods, since
                        # multiplying left and right by -1 yields same result.
                        left = ro_Q @ core_U[:, r:]
                        right = core_Vt[r:, :] @ lo_Qt

                        for idx in range(svectors_check_n):
                            recons_err = torch.dist(
                                torch.outer(left[:, idx], right[idx]),
                                torch.outer(U[:, idx + r], Vt[idx + r]),
                            ) / (U[:, idx + r].norm() * Vt[idx + r].norm())
                            assert recons_err <= abs(
                                svectors_frob_tol[dtype]
                            ), "Bad recovery of singular vectors!"
                        # check a-posteriori error. With 30 measurements, we
                        # know that the probability of the actual error being
                        # twice as big is less than 1%, and keeps shrinking
                        # exponentially: a_posteriori_error_bounds(30, 1)
                        # Therefore, a very small estimate almost certainly
                        # guarantees that the approximation is good, and we
                        # have enough rank in our truncated SVD.
                        for adjoint in [True, False]:
                            _, _, err_estimate = a_posteriori_error(
                                mat,
                                mat_recons,
                                post,
                                seed=seed,
                                dtype=dtype,
                                device=device,
                                adjoint=adjoint,
                            )[0]
                            assert (
                                err_estimate / mat.numel()
                            ) <= atol, "Bad a-posteriori estimate!"
                        # cleanup HDF5 tmpdir
                        tmpdir.cleanup()


# ##############################################################################
# # SKETCHED EIGH
# ##############################################################################
def test_dseigh_exp(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    atol_exp,
    dims_ranks_outer_inner_posteriori,
    steep_exp_decay,
    recons_frob_tol,
    svectors_frob_tol,
    svectors_check_n,
):
    """Test distributed SEIGH on (PSD and non-PSD) exponential matrices.

    Sample matrices, and compute their ``torch.eigh`` as well as the sketched
    EIGH. Test that:
    * Recovered eigenvalues are correct
    * Recovered matrix is correct and symmetric
    * The first bulk eigenvectors are correct
    * A-posteriori error estimation is low
    """
    for seed in rng_seeds:
        for dtype, atol in atol_exp.items():
            for d, r, (outer, inner), post in dims_ranks_outer_inner_posteriori:
                for dec in steep_exp_decay:
                    for psd in (True, False):
                        # we create matrix and full eigh on cpu, it is faster
                        mat = SynthMat.exp_decay(
                            shape=(d, d),
                            rank=r,
                            decay=dec,
                            symmetric=True,
                            seed=seed,
                            dtype=dtype,
                            device="cpu",
                            psd=psd,
                        )
                        S, U = torch.linalg.eigh(mat)  # mat = U @ diag(S) @ U.T
                        # sort eigdec in descending eigval magnitude:
                        _, perm = abs(S).sort(descending=True)
                        S, U = S[perm], U[:, perm]
                        # make sure we actually have negative eigvals
                        if not psd:
                            assert (S < -abs(atol)).any(), (
                                "non-PSD should have " + "negative eigvals!"
                            )
                        # then we test decomposition on both devices
                        for device in torch_devices:
                            mat = mat.to(device)
                            S = S.to(device)
                            U = U.to(device)
                            #
                            tmpdir = tempfile.TemporaryDirectory()
                            (Q, core_U, core_S) = mock_dseigh(
                                tmpdir.name,
                                mat,
                                device,
                                dtype,
                                outer,
                                inner,
                                seed,
                            )
                            # check that singular values are correct
                            assert torch.allclose(
                                S[: 2 * r], core_S[: 2 * r], atol=atol
                            ), "Bad recovery of singular values!"

                            # check that recovered matrix is the same
                            core_U, core_S = truncate_core(
                                2 * r, core_U, core_S
                            )
                            mat_recons = (
                                Q @ core_U @ torch.diag(core_S) @ core_U.T @ Q.T
                            )
                            # check that reconstruction is symmetric
                            assert torch.allclose(
                                mat_recons, mat_recons.T, atol=atol
                            ), "Recovered sseigh is not symmetric!"
                            recons_err = (
                                torch.dist(mat, mat_recons) / mat.norm()
                            )
                            assert recons_err <= abs(
                                recons_frob_tol[dtype]
                            ), f"Bad SSVD reconstruction! {recons_err}"

                            # check a few singular vectors are correct: avoid
                            # the ones that have the same singular values, since
                            # they aren't unique. And compare outer prods, since
                            # multiplying both sides by -1 yields same result.
                            recons_evs = Q @ core_U[:, r:]
                            for idx in range(svectors_check_n):
                                recons_err = torch.dist(
                                    recons_evs[:, idx], U[:, r + idx]
                                ).item()
                                recons_err_flip = torch.dist(
                                    recons_evs[:, idx], -U[:, r + idx]
                                ).item()
                                recons_err = min(recons_err, recons_err_flip)
                                assert recons_err <= abs(
                                    svectors_frob_tol[dtype]
                                ), "Bad recovery of singular vectors!"
                            # check a-posteriori error. With 30 measurements, we
                            # know that the prob of the actual error being
                            # twice as big is less than 1%, and keeps shrinking
                            # exponentially: a_posteriori_error_bounds(30, 1)
                            # Therefore, a very small estimate almost certainly
                            # guarantees that the approximation is good, and we
                            # have enough rank in our truncated SVD.
                            for adjoint in [True, False]:
                                _, _, err_estimate = a_posteriori_error(
                                    mat,
                                    mat_recons,
                                    post,
                                    seed=seed,
                                    dtype=dtype,
                                    device=device,
                                    adjoint=adjoint,
                                )[0]
                                assert (
                                    err_estimate / mat.numel()
                                ) <= atol, "Bad a-posteriori estimate!"
                            # cleanup HDF5 tmpdir
                            tmpdir.cleanup()


def test_dseigh_poly(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    atol_poly,
    dims_ranks_outer_inner_posteriori,
    steep_poly_decay,
    recons_frob_tol,
    svectors_frob_tol,
    svectors_check_n,
):
    """Test distributed SEIGH on (PSD and non-PSD) polynomial matrices.

    Sample matrices, and compute their ``torch.eigh`` as well as the sketched
    EIGH. Test that:
    * Recovered eigenvalues are correct
    * Recovered matrix is correct and symmetric
    * The first bulk eigenvectors are correct
    * A-posteriori error estimation is low
    """
    for seed in rng_seeds:
        for dtype, atol in atol_poly.items():
            for d, r, (outer, inner), post in dims_ranks_outer_inner_posteriori:
                for dec in steep_poly_decay:
                    for psd in (True, False):
                        # we create matrix and full eigh on cpu, it is faster
                        mat = SynthMat.poly_decay(
                            shape=(d, d),
                            rank=r,
                            decay=dec,
                            symmetric=True,
                            seed=seed,
                            dtype=dtype,
                            device="cpu",
                            psd=psd,
                        )
                        S, U = torch.linalg.eigh(mat)  # mat = U @ diag(S) @ U.T
                        # sort eigdec in descending eigval magnitude:
                        _, perm = abs(S).sort(descending=True)
                        S, U = S[perm], U[:, perm]
                        # make sure we actually have negative eigvals
                        if not psd:
                            assert (S < -abs(atol)).any(), (
                                "non-PSD should have " + "negative eigvals!"
                            )
                        # then we test decomposition on both devices
                        for device in torch_devices:
                            mat = mat.to(device)
                            S = S.to(device)
                            U = U.to(device)
                            #
                            tmpdir = tempfile.TemporaryDirectory()
                            (Q, core_U, core_S) = mock_dseigh(
                                tmpdir.name,
                                mat,
                                device,
                                dtype,
                                outer,
                                inner,
                                seed,
                            )
                            # check that singular values are correct
                            assert torch.allclose(
                                S[: 2 * r], core_S[: 2 * r], atol=atol
                            ), "Bad recovery of singular values!"

                            # check that recovered matrix is the same
                            core_U, core_S = truncate_core(
                                2 * r, core_U, core_S
                            )
                            mat_recons = (
                                Q @ core_U @ torch.diag(core_S) @ core_U.T @ Q.T
                            )
                            # check that reconstruction is symmetric
                            assert torch.allclose(
                                mat_recons, mat_recons.T, atol=atol
                            ), "Recovered sseigh is not symmetric!"
                            recons_err = (
                                torch.dist(mat, mat_recons) / mat.norm()
                            )
                            assert recons_err <= abs(
                                recons_frob_tol[dtype]
                            ), f"Bad SSVD reconstruction! {recons_err}"

                            # check a few singular vectors are correct: avoid
                            # the ones that have the same singular values, since
                            # they aren't unique. And compare outer prods, since
                            # multiplying both sides by -1 yields same result.
                            recons_evs = Q @ core_U[:, r:]
                            for idx in range(svectors_check_n):
                                recons_err = torch.dist(
                                    recons_evs[:, idx], U[:, r + idx]
                                ).item()
                                recons_err_flip = torch.dist(
                                    recons_evs[:, idx], -U[:, r + idx]
                                ).item()
                                recons_err = min(recons_err, recons_err_flip)
                                assert recons_err <= abs(
                                    svectors_frob_tol[dtype]
                                ), "Bad recovery of singular vectors!"
                            # check a-posteriori error. With 30 measurements, we
                            # know that the prob of the actual error being
                            # twice as big is less than 1%, and keeps shrinking
                            # exponentially: a_posteriori_error_bounds(30, 1)
                            # Therefore, a very small estimate almost certainly
                            # guarantees that the approximation is good, and we
                            # have enough rank in our truncated SVD.
                            for adjoint in [True, False]:
                                _, _, err_estimate = a_posteriori_error(
                                    mat,
                                    mat_recons,
                                    post,
                                    seed=seed,
                                    dtype=dtype,
                                    device=device,
                                    adjoint=adjoint,
                                )[0]
                                assert (
                                    err_estimate / mat.numel()
                                ) <= atol, "Bad a-posteriori estimate!"
                            # cleanup HDF5 tmpdir
                            tmpdir.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest configuration.

This configuration module is implicitly parsed by ``pytest``.
"""


import pytest


# ##############################################################################
# # CLI ARGUMENTS
# ##############################################################################
def str_of_unique_ints(x):
    """Typechecker function for pytest CLI arguments.

    Input ``x`` is expected to be a string with integers in the form '0 1 2 3'.
    If that is the case, returns a list ``[0, 1, 2, 3]``. Otherwise, a
    ``pytest.UsageError`` is raised.
    """
    # handle the empty case
    if isinstance(x, str) and len(x) == 0:
        return []
    # otherwise, test for list of unique integers
    sep = " "
    try:
        result = [int(elt) for elt in x.split(sep)]
        assert len(result) == len(set(result))  # repeated values?
    except (ValueError, AssertionError) as err:
        raise pytest.UsageError(
            "cmdopt must be a string with unique spase-separated integers!"
        ) from err
    #
    return result


def pytest_addoption(parser):
    """Adds CLI option to the ``pytest`` command.

    Usage example::

      pytest ... --seeds='1 2 3 4'
    """
    parser.addoption(
        "--seeds",
        action="store",
        default="0 1 -12345 12345 479915",  # 0b1110101001010101011
        help="Space-separated string of integers to use as seeds.",
        type=str_of_unique_ints,
    )
    parser.addoption(
        "--quick",
        action="store_true",
        help="If given, less tests will be run",
    )
    parser.addoption(
        "--skip_toomanyfiles",
        action="store_true",
        help="If given, test_too_many_files will do nothing (but still run).",
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Pytest for in-core sketched (sub-)diagonal estimation."""


import pytest
import torch

from skerch.subdiagonals import subdiag_hadamard_pattern, subdiagpp
from skerch.synthmat import SynthMat
from skerch.utils import gaussian_noise

from . import rng_seeds, torch_devices  # noqa: F401


# ##############################################################################
# # FIXTURES
# ##############################################################################
@pytest.fixture
def dim_rank_decay_sym_diags_meas_defl_rtol(request):
    """Test cases for main diagonal estimation on different matrices.

    Entries are in the form
    ``(dim, rank, decay, sym, [diag_idxs], num_meas, defl, rtol)``, where:

    * ``dim`` is the number of dimensions of a square test matrix of
      exp-decaying singular values
    * ``rank`` is the number of unit singular values in the test matrix
    * ``decay`` is the speed of the exp-decay of singular values after ``rank``.
      The larger, the faster decay.
    * ``sym`` is a boolean specifying whether the matrix is symmetric
    * ``diag_idxs`` is a collection of tuples of integers specifying the
      diagonal indices to be tested: 0 is the main diagonal, 1 the diagonal
      above, -1 the diagonal below, and so on.
    * ``num_meas`` specifies the number of measurements for the Hutchinson
      diagonal estimator
    * ``defl`` specifies the rank of the orthogonal projector for the
      Hutchinson++ deflation.

    .. note::

      The given diagonal indices should not exceed the furthest away
      diagonal, which depends on the shape.

    This is a bit of a complicated fixture, the reason being that the tolerance
    and number of measurements needed depend heavily on the type of matrix
    encountered: symmetric random matrices have strong diagonals, which lead
    to much faster Hutchinson convergence. Low-rank matrices can be effectively
    characterized via deflation, which also helps. And so on.
    """
    dims, rank = 1000, 100
    if request.config.getoption("--quick"):
        dims, rank = 500, 50
    result = [
        # fast-decay: just Hutch does poorly, but better if symmetric
        (dims, rank, 0.5, True, [0], round(dims * 0.995), 0, 0.02),
        (dims, rank, 0.5, False, [0], round(dims * 0.995), 0, 0.1),
        # slow-decay: just Hutch behaves a bit worse as with fast decay
        (dims, rank, 0.01, True, [0], round(dims * 0.995), 0, 0.02),
        (dims, rank, 0.01, False, [0], round(dims * 0.995), 0, 0.1),
        # fast-decay: just deflating is great (also for asymmetric)
        (dims, rank, 0.5, True, [0], 0, rank + 10, 1e-5),
        (dims, rank, 0.5, False, [0], 0, rank + 10, 1e-4),
        # slow-decay: deflating is less good and affected by asym
        (dims, rank, 0.01, True, [0], 0, rank * 3, 0.05),
        (dims, rank, 0.01, False, [0], 0, rank * 4, 0.06),
        # slow-decay: A lot of Hutch are needed, but deflation tends to help
        # for asym
        (dims, rank, 0.01, True, [0], round(dims * 0.7), rank * 3, 0.02),
        (dims, rank, 0.01, False, [0], round(dims * 0.7), rank * 4, 0.04),
    ]
    return result


@pytest.fixture
def dim_rank_decay_sym_subdiags_meas_defl_rtol(request):
    """Test cases for subdiagonal estimation on different matrices.

    Note that very extremal subdiagonals have few entries, and we are better
    off just directly measuring them. For this reason, this test focuses on
    subdiagonals in the mid-section. It also focuses on asymmetric matrices
    only, since symmetry mostly impacts the main diagonal.
    """
    dims, rank = 100, 10
    if request.config.getoption("--quick"):
        subdiag_idxs = [-50, -40, -30, -20, -10, 10, 20, 30, 40, 50]
    else:
        subdiag_idxs = [-30, -10, 10, 30]
    #
    result = [
        # fast-decay: low-rank recovery works nicely
        (dims, rank, 0.5, False, subdiag_idxs, 0, rank * 2, 5e-5),
        # mid-decay: low-rank recovery works less nicely for subdiags
        (dims, rank, 0.1, False, subdiag_idxs, 0, rank * 3, 0.03),
        # fast-decay: Hutch is bad for central subdiags (worse than direct)...
        (dims, rank, 0.5, False, subdiag_idxs, 99, 0, 0.2),
        # but as with main diag, Hutch doesn't depend much on the decay
        (dims, rank, 0.1, False, subdiag_idxs, 99, 0, 0.2),
    ]
    return result


@pytest.fixture
def subdiag_hadamard_idxs():
    """Subdiagonal indices to test with the Hadamard pattern."""
    result = [
        [0, 1, 2, 3, 4, 5],
        [0, -1, -2, -3, -4, -5],
        [-1, 5, -9, 9],
    ]
    return result


@pytest.fixture
def hadamard_atol():
    """Absolute tolerances to test Hadamard pattern."""
    result = {torch.float64: 1e-13, torch.float32: 1e-5}
    return result


# ##############################################################################
# # HADAMARD
# ##############################################################################
def test_subdiag_hadamard_pattern(
    rng_seeds,  # noqa: F811
    hadamard_atol,
    torch_devices,  # noqa: F811
    subdiag_hadamard_idxs,
):
    """Test case for generic Hadamard pattern generator.

    Given a vector and a set of indices, the Hadamard pattern is a
    non-circular convolution using indices as deltas in the impulse response.
    This pattern is useful to estimate sub-diagonals using Hutchinson.

    This test case generates random vectors and checks that the responses
    are right for different indices. It also checks that the FFTconv
    implementation and the plain one yield same results.
    """
    for seed in rng_seeds:
        for dtype, atol in hadamard_atol.items():
            for device in torch_devices:
                # Simple tests with delta responses:
                v = gaussian_noise(100, seed=seed, dtype=dtype, device=device)
                had = subdiag_hadamard_pattern(v, [], use_fft=False)
                assert had.norm() == 0, "Empty idxs should yield zeros!"
                had = subdiag_hadamard_pattern(v, [0], use_fft=False)
                assert torch.allclose(
                    v, had, atol=atol
                ), "Delta(0) should yield same result!"
                had = subdiag_hadamard_pattern(v, [1], use_fft=False)
                assert (had[0] == 0) and (
                    torch.allclose(had[1:], v[:-1], atol=atol)
                ), "Delta(1) should shift vector one position to the right!"
                had = subdiag_hadamard_pattern(v, [-1], use_fft=False)
                assert (had[-1] == 0) and (
                    torch.allclose(had[:-1], v[1:], atol=atol)
                ), "Delta(-1) should shift vector one position to the left!"
                # random test with more complex responses
                for idxs in subdiag_hadamard_idxs:
                    v = gaussian_noise(
                        100,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                    )
                    had = subdiag_hadamard_pattern(v, idxs, use_fft=False)
                    had_fft = subdiag_hadamard_pattern(v, idxs, use_fft=True)
                    assert torch.allclose(
                        had, had_fft, atol=atol
                    ), "Inconsistent FFT implementation of Hadamard pattern!"
                    if not idxs:
                        assert had.norm() == 0, "Hadamard should be empty!"


# ##############################################################################
# # DIAG ESTIMATORS
# ##############################################################################
def test_main_diags(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    dim_rank_decay_sym_diags_meas_defl_rtol,
):
    """Test case for main diagonal estimation.

    This test creates an ``exp_decay`` random (square) test matrix, and checks
    that its main diagonal is estimated within ``rtol``.

    Depending on the passed fixture, the matrix will have different order,
    symmetry and spectral decay.
    """
    for seed in rng_seeds:
        for dtype in (torch.float64, torch.float32):
            for device in torch_devices:
                for (
                    dim,
                    rank,
                    decay,
                    sym,
                    diags,
                    meas,
                    defl,
                    rtol,
                ) in dim_rank_decay_sym_diags_meas_defl_rtol:
                    mat = SynthMat.exp_decay(
                        shape=(dim, dim),
                        rank=rank,
                        decay=decay,
                        symmetric=sym,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                        psd=False,
                    )
                    for diag_idx in diags:
                        # retrieve the true diag
                        diag = torch.diag(mat, diagonal=diag_idx)
                        # matrix-free estimation of the diag
                        diag_est, _, norms = subdiagpp(
                            mat,
                            meas,
                            dtype,
                            device,
                            seed + 1,
                            defl,
                            diag_idx,
                        )
                        # then assert
                        dist = torch.dist(diag, diag_est)
                        rel_err = (dist / torch.norm(diag)).item()
                        assert rel_err <= rtol, "Incorrect diagonal recovery!"


def test_subdiags(
    rng_seeds,  # noqa: F811
    torch_devices,  # noqa: F811
    dim_rank_decay_sym_subdiags_meas_defl_rtol,
):
    """Test case for subdiagonal estimation.

    This test creates an ``exp_decay`` random test matrix, and checks that
    several of its subdiagonals are estimated within ``rtol``.

    """
    for seed in rng_seeds:
        for dtype in (torch.float64, torch.float32):
            for device in torch_devices:
                for (
                    dim,
                    rank,
                    decay,
                    sym,
                    diags,
                    meas,
                    defl,
                    rtol,
                ) in dim_rank_decay_sym_subdiags_meas_defl_rtol:
                    mat = SynthMat.exp_decay(
                        shape=(dim, dim),
                        rank=rank,
                        decay=decay,
                        symmetric=sym,
                        seed=seed,
                        dtype=dtype,
                        device=device,
                        psd=False,
                    )
                    for diag_idx in diags:
                        # retrieve the true diag
                        diag = torch.diag(mat, diagonal=diag_idx)
                        # matrix-free estimation of the diag
                        diag_est, _, norms = subdiagpp(
                            mat,
                            meas,
                            dtype,
                            device,
                            seed + 1,
                            defl,
                            diag_idx,
                        )
                        # then assert
                        dist = torch.dist(diag, diag_est)
                        rel_err = (dist / torch.norm(diag)).item()
                        assert rel_err <= rtol, "Incorrect subdiag recovery!"



# -*- coding: utf-8 -*-

r"""Sketched (Sub-)Diagonals
============================

In this tutorial we create a noisy, numerically low-rank matrix,
and perform sketched estimations of its main diagonal and subdiagonals.
"""

import matplotlib.pyplot as plt
import torch

from skerch.subdiagonals import subdiagpp
from skerch.synthmat import SynthMat

# %%
#
# ##############################################################################
#
# Globals and creation of test matrix
# -----------------------------------
#
# We will create a random matrix of shape ``(ORDER, ORDER)``, with an effective
# ``RANK`` of unit singular values followed by exponential decay with rate
# ``EXP_DECAY``.

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DTYPE = torch.float64
ORDER, RANK = 1000, 100
EXP_DECAY = 0.2
SEED = 12345


linop = SynthMat.exp_decay(
    shape=(ORDER, ORDER),
    rank=RANK,
    decay=EXP_DECAY,
    symmetric=False,
    seed=SEED,
    dtype=DTYPE,
    device=DEVICE,
    psd=False,
)

# %%
#
# ##############################################################################
#
# Exact (sub-)diagonals
# ---------------------
#
# For testing purposes, we extract the exact main diagonal, as well as the
# sub-diagonals right above and below

main_diag = torch.diag(linop, diagonal=0)
sub_diag = torch.diag(linop, diagonal=-1)
super_diag = torch.diag(linop, diagonal=1)

print("Norm of main diagonal:", main_diag.norm().item())
print("Norm of lower subdiagonal:", sub_diag.norm().item())
print("Norm of upper subdiagonal:", super_diag.norm().item())

# %%
#
# ##############################################################################
#
# Sketched approximation of main diagonal
# ---------------------------------------
#
# for this initial approximation we do not deflate, and just rely on the
# Hutchinson estimator:

main_diag_est, _, (top_norm, bottom_norm) = subdiagpp(
    linop,
    int(ORDER * 0.5),
    DTYPE,
    DEVICE,
    SEED + 1,
    0,
    0,
)

fig, ax = plt.subplots()
ax.plot(main_diag.cpu()[:100], color="black", label="original")
ax.plot(main_diag_est.cpu()[:100], color="pink", label="approximation")
ax.legend()
fig.tight_layout()

print(
    "Norm of residual (main diagonal):",
    torch.dist(main_diag, main_diag_est).item(),
)

# %%
#
# Since we are not deflating, we confirm that all of the obtained diagonal
# comes from the deflated component

print("Norm of top-rank component:", top_norm)
print("Norm of deflated component:", bottom_norm)

# %%
#
# ##############################################################################
#
# Sketched approximation of subdiagonals
# --------------------------------------
#
# Now we do the same for the neighboring subdiagonals

sub_diag_est, _, (top_norm, bottom_norm) = subdiagpp(
    linop,
    int(ORDER * 0.5),
    DTYPE,
    DEVICE,
    SEED + 1,
    0,
    -1,
)

super_diag_est, _, (top_norm, bottom_norm) = subdiagpp(
    linop,
    int(ORDER * 0.5),
    DTYPE,
    DEVICE,
    SEED + 1,
    0,
    1,
)

fig, (ax1, ax2) = plt.subplots(nrows=2)
ax1.plot(super_diag.cpu()[:100], color="black")
ax1.plot(super_diag_est.cpu()[:100], color="pink", alpha=0.8)
ax2.plot(sub_diag.cpu()[:100], color="black")
ax2.plot(sub_diag_est.cpu()[:100], color="pink", alpha=0.8)
fig.tight_layout()

print(
    "Norm of residual (upper subdiagonal):",
    torch.dist(super_diag, super_diag_est).item(),
)
print(
    "Norm of residual (lower subdiagonal):",
    torch.dist(sub_diag, sub_diag_est).item(),
)

# %%
#
# ##############################################################################
#
# Deflation
# ---------
#
# Since our test matrix can be well-approximated with low rank, deflation can
# help a lot. Here we combine deflation with Hutchinson measurements, and show
# that the recovery is much better, and that most of it is due to the deflation.

main_diag_est, defl_mat, (top_norm, bottom_norm) = subdiagpp(
    linop,
    int(ORDER * 0.5),
    DTYPE,
    DEVICE,
    SEED + 1,
    RANK,
    0,
)

fig, ax = plt.subplots()
ax.plot(main_diag.cpu()[:100], color="black")
ax.plot(main_diag_est.cpu()[:100], color="pink")
fig.tight_layout()

print(
    "Norm of residual (main diagonal):",
    torch.dist(main_diag, main_diag_est).item(),
)
print("Shape of deflation matrix:", tuple(defl_mat.shape))
print("Norm of top-rank component:", top_norm.item())
print("Norm of deflated component:", bottom_norm)
